#!/usr/local/bin/Rscript
#
# This R script is provided as a Supplemental code to the paper:
#   Louca and Pennell (in review as of 2021). Why extinction estimates from extant phylogenies are so often zero.
#
# Included are:
#	Fitting ELC birth-death models (BDELC) to timetrees simulated under time-dependent speciation and extinction rates, while either constraining the extinction rate to non-negative values or allowing for negative values (BDECNeg)
#	Fitting BDELC and BDELCNeg models to a collection of empirical timetrees
#	Repeating the simulation analysis with slightly erroneous trees (see option SIM_NOISY)
#
# To run this script, type the following in your terminal:
#	Rscript workflow.R
# The script will save all output in a newly generated output folder, each time it is ran (i.e., previous output is not replaced).
#
# Tested on R 4.0.2, MacOS 10.13.6, castor v1.6.7
# On a 2015 MacBook Pro, the full script takes about 7 hours to finish.
# Tip: To change the number of simulations (and thus computation time), modify the variable SIM_FITTING_NSIMULATIONS.
#
# The script requires the package castor v1.6.7 or later, which can be downloaded from:
#   www.loucalab.com/archive/castor
# Note that the script will not work with older versions of castor!
#
# LICENSE AGREEMENT
# - - - - - - - - -
# Use and redistributions of this code is permitted, provided that the following 
# conditions are always met:
#
#    * Redistributions must retain the above copyright notice, this list of 
#      conditions and the following disclaimer in the code itself, as well 
#      as in documentation and/or other materials provided with the code.
#    * Neither the name of the original author (Stilianos Louca), nor the names 
#      of its contributors may be used to endorse or promote products derived 
#      from this code without specific prior written permission.
#
# THIS CODE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS CODE, 
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# - - - - - - - - -
#
# Stilianos Louca
# December 19, 2020


###################################
# OPTIONS

REQUIRED_PACKAGES=c("castor", "stringr")

OUTPUT_DIR="output"
TREE_SPECS_FILE_PATH="input/tree_descriptions.tsv"
OMIT_TREES = "" # optional comma-separated list of tree names (or wildcards) to omit from analysis
ONLY_TREES = "" # optional comma-separated list of tree names to restrict analysis to

NUMBER_OF_PARALLEL_THREADS	= 4
INTEGRATION_RELATIVE_DT 	= 1e-3
MODEL_ADEQUACY_NBOOTSTRAPS 	= 1000 # number of bootstraps used to assess model adequacy (statistical significance of verious tree statistics)

# general fitting options
fitting_Ntips2max_model_runtime = function(Ntips) max(2,Ntips/1e4) # runtime in seconds to allocate for likelihood evaluations during fitting, as a function of tree size
FITTING_NTRIALS				= 1
FITTING_NITERATIONS 		= 1000
FITTING_NEVALUATIONS 		= 1000
FITTING_REL_TOLERANCE 		= 1e-12
FITTING_STEP_MIN			= 0.00001

# options specifically for fitting models to real trees
REAL_FITTING_MIN_LTT						= 0
REAL_FITTIN_MIN_NTIPS					 	= 10	 	# don't analyze any trees with fewer tips than this
REAL_FITTING_MIN_RAREFACTION				= 0.001 	# don't analyze any trees with lower sampling fraction than this
REAL_FITTING_BDELCNEG_USE_BDELC_MODEL 	 	= TRUE 		# use the best ELC model to fit the model with negative mu, instead of considering all ELC models with negative mu
REAL_FITTING_BDELCNEG_USE_BDELC_AS_GUESS  	= FALSE		# use the fitted rate profiles from BDELC as starting point for fitting BDELCNeg, if available.

# options for fitting models to simulated trees
SIM_FITTING_AGE_FRACTION					= 1
SIM_FITTING_NSIMULATIONS 					= 100
SIM_FITTING_REPEAT_FAILED_TREES				= TRUE
SIM_FITTING_BDELCNEG_USE_BDELC_MODEL		= TRUE # use the best ELC model to fit the model with negative mu, instead of considering all ELC models with negative mu
SIM_FITTING_BDELCNEG_USE_BDELC_AS_GUESS		= FALSE

SIM_TIME_UNIT = "Myr" # time unit to assume for simulated trees
SIM_BDVAR_TREE_SIZE=1000

# options for adding noise to simulated trees
SIM_NOISY							= FALSE	# if TRUE, then noise will be added to simulated trees, by randomly modifying node ages and merging a random subset of nodes into multifurcations
NOISY_SIM_MULTIFURCATION_FRACTION	= 0.2 	# fraction of nodes to be merged upward with their parents, to create multifurcations
NOISY_SIM_MAX_RELATIVE_AGE_ERROR	= 0.2 	# maximum random age error added to nodes, relative to their true age

# which analyses to perform
INCLUDE_BDELC_FITTING_TO_SIMULATIONS	= TRUE
INCLUDE_MODEL_FITTING_TO_REAL_TREES		= TRUE

# enable or disable certain types of model fitting to real trees
REAL_TREES_INCLUDE_HBD_ELC_FITTING 		= TRUE
REAL_TREES_INCLUDE_HBD_ELC_NEG_FITTING	= TRUE

# plot styles for specific data types
PLOT_COLOR_FITTED		= "#1c568e"
PLOT_COLOR_FITTED2		= "#711b8e"
PLOT_COLOR_EMPIRICAL	= "#303030"
PLOT_LINE_TYPE_EMPIRICAL= 1
PLOT_LINE_TYPE_FITTED	= 2
PLOT_LINE_TYPE_FITTED2	= 3
BLACK_CURVE_COLOR		= "#303030"
BLUE_CURVE_COLOR		= "#005a96"
CI50_CURVE_COLOR		= "#808080"
CI50_SHADE_COLOR		= "#A0A0A0"
CI95_CURVE_COLOR		= "#BBBBBB"
CI95_SHADE_COLOR		= "#DDDDDD"
REFERENCE_CURVE_COLOR	= "#606060"

# default plot style palettes
PLOT_COLOR_PALETTE=c(BLACK_CURVE_COLOR, BLUE_CURVE_COLOR, "red", "brown", "darkgreen", "#D3762A", "#618ACC", "#B80DC4", "#565656")
PLOT_LINE_TYPE_PALETTE=rep(c(1,1,2,2,3,3),ages=5)
PLOT_LINE_WIDTH_PALETTE=rep(c(2,1),ages=5)

DEFAULT_PLOT_WIDTH						= 2 	# plot width in inches
DEFAULT_PLOT_HEIGHT						= 2  	# plot height in inches
DEFAULT_PLOT_TOP_MARGIN					= 1.5
DEFAULT_PLOT_RIGHT_MARGIN				= 0.5
DEFAULT_PLOT_RIGHT_MARGIN_WITH_LEGEND 	= 3 # inches
PLOT_DOWNSAMPLING_RESOLUTION			= 10000	# downsample plotted time series to this many time points if needed, for efficiency


options(expressions=100000) # increase recursion depth
options(max.print=1000000) # increase print count
options(warn=1) # print warnings as they occur


###################################
# AUXILIARY FUNCTIONS


check_output_file = function(file_path,force_replace,verbose,verbose_prefix){
	if(file.exists(file_path)){
		if(force_replace){
			cat(sprintf("%sNote: Replacing output file '%s'.\n",verbose_prefix,file_path))
			file.remove(file_path);
		}else{
			stop(sprintf("Output file '%s' already exists. Cowardly refusing to continue.",file_path), call.=FALSE)
		}
	}
	dir.create(dirname(file_path), showWarnings = FALSE, recursive=TRUE);
}


get_non_existent_dir = function(parent_path, child_basename, digits=3){
	existing = list.files(path=parent_path, full.names=FALSE, recursive=FALSE)
	counter = 1
	child_name = sprintf(sprintf("%%s%%0%dd",digits),child_basename,counter)
	while(any(sapply(existing, FUN=function(x) startsWith(x,child_name)))){
		counter = counter + 1;
		child_name = sprintf(sprintf("%%s%%0%dd",digits),child_basename,counter)
	}
	child_path = file.path(parent_path,child_name)
	return(child_path);
}


file_basename = function(filepath){
	filebasename = basename(filepath)
	# remove extension
	parts = strsplit(filebasename, split=".", fixed=TRUE)[[1]]
	if(length(parts)==1){
		return(filebasename)
	}else{
		return(paste(parts[1:(length(parts)-1)], collapse="."))
	}
}


file_basepath = function(filepath){
	# remove extension
	parts = strsplit(filepath, split=".", fixed=TRUE)[[1]]
	if(length(parts)==1){
		return(filepath)
	}else{
		return(paste(parts[1:(length(parts)-1)], collapse="."))
	}
}


# improved version of approx(), able to handle situations where the input curve only has one grid point
approx1 = function(	x,
					y,
					xout,
					rule=1){
	if(length(x)==0) return(list(y=rep(NA,length(xout))))
	if(length(x)==1) return(list(y=rep(y,length(xout))))
	return(approx(x=x,y=y,xout=xout,rule=rule))
						
}


# calculate the pulled diversification rate (PDR) based on lambda & mu
get_PDR_on_grid = function(age_grid, lambdas, mus){
	NG				= length(age_grid)
	lambda_slopes 	= diff(lambdas)/diff(age_grid);
	lambda_slopes 	= c(lambda_slopes[1],lambda_slopes,tail(lambda_slopes,1)); # add dummy slope to the left & right, so that lambda_slopes becomes of length NG+1
	lambda_slopes 	= 0.5*(lambda_slopes[1:NG]+lambda_slopes[2:(NG+1)]);
	PDRs			= lambdas - mus + lambda_slopes/lambdas;
	return(PDRs);
}



filter_items_by_name = function(item_names, 	# character vector of names
								only_names, 	# (character) comma-separated list of names or wildcards
								omit_names){	# (character) comma-separated list of names or wildcards
	N = length(item_names)
	if(only_names!=""){
		keep = rep(FALSE, times=N)
		only_names = sapply(strsplit(only_names,",")[[1]],utils::glob2rx)
		for(w in 1:length(only_names)){
			keep = (keep | grepl(only_names[w], item_names, perl=TRUE))
		}
	}
	else{
		keep = rep(TRUE, times=N)
	}
	if(omit_names!=""){
		omit_names = sapply(strsplit(omit_names,",")[[1]],utils::glob2rx)
		for(w in 1:length(omit_names)){
			keep = (keep & (!grepl(omit_names[w], item_names, perl=TRUE)))
		}
	}
	return(which(keep))
}


add_noise_to_tree = function(tree, multifurcation_fraction, max_relative_age_error){
	Ntips		= length(tree$tip.label)
	Nnodes		= tree$Nnode
	
	# randomly modify node ages
	clade_times 	= castor::get_all_distances_to_root(tree)
	root_age		= max(clade_times)
	clade_ages		= root_age-clade_times
	node_shifts		= runif(n=Nnodes, min=-max_relative_age_error, max=max_relative_age_error) * clade_ages[(Ntips+1):(Ntips+Nnodes)]
	tree = castor::shift_clade_times(tree, clades_to_shift=((Ntips+1):(Ntips+Nnodes)), time_shifts=node_shifts, shift_descendants = FALSE, negative_edge_lengths = "move_parent")$tree

	# merge a random subset of nodes into multifurcations
	nodes_to_merge = sample.int(n=Nnodes, size=as.integer(multifurcation_fraction*Nnodes), replace=FALSE)
	tree = castor::merge_nodes_to_multifurcations(tree, nodes_to_merge=nodes_to_merge, merge_with_parents=FALSE, keep_ancestral_ages=FALSE)$tree
	
	return(tree)
}



get_R2 = function(xtrue, ytrue, xfit, yfit, minx=NULL, maxx=NULL){
	if(is.null(minx)) minx = min(xtrue)
	if(is.null(maxx)) maxx = max(xtrue)
	yfit_on_xtrue 	= suppressWarnings(approx(x=xfit, y=yfit, xout=xtrue)$y)
	valids 			= which(is.finite(yfit_on_xtrue) & is.finite(ytrue) & (xtrue>=minx) & (xtrue<=maxx)) # only consider points where the fitted model and the data are defined
	if(length(valids)<=1) return(NaN);
	ytrue 			= ytrue[valids]
	xtrue 			= xtrue[valids]
	yfit_on_xtrue 	= yfit_on_xtrue[valids]
	SSR 			= sum((yfit_on_xtrue-ytrue)**2)
	return(1 - SSR/(length(ytrue)*var(ytrue)))
}



# save & plot variables over age
plot_curves = function(	file_basepath, 		# e.g. 'output/SILVA_curves_last_1000years'
						xtype,				# e.g. 'age'
						ytype,				# e.g. 'birth_rates'
						case_tag,			# e.g. 'model 2'
						curves, 			# list of size Ncurves, each entry of which is a sub-list with two elements (ages, values) specifying a separate curve to be plotted
						curve_names,		# 1D character vector of size >=Ncurves
						curve_colors		= NULL,		# either NULL or a 1D vector of size >=Ncurves, specifying the color of a curve. If NULL, colors are picked automatically.
						curve_line_types	= NULL,			# either NULL or a 1D vector of size >=Ncurves, specifying the line type of a curve. If NULL, line types are picked automatically.
						curve_widths		= NULL,			# either NULL or a 1D vector of size >=Ncurves, specifying the line width of a curve. If NULL, line widths are picked automatically.
						show_curves			= NULL,			# either a 1D vector of booleans of size Ncurves, indicating whether a curve should be plotted, or NULL (plot all curves)
						shadings			= NULL,			# optional list of shading specifications. Each specification is a list of 3 entries (the 2 curves to shade between and the shading color). Can also be NULL. For example, list(list(1,2,'red'),list(5,6,'blue')) will shade the area between curves 1 & 2 in red and the area between curves 5 & 6 in blue.
						reverse_x			= FALSE,		# boolean, specifying if the x-axis should be plotted reversed
						plot_minx			= NULL,			# minimum x-value to plot. If NULL, this is automatically set by the data
						plot_maxx			= NULL,			# maximum x-value to plot. If NULL, this is automatically set by the data
						plot_miny			= NULL,			# optional lower bound for the plots. If NULL, this is automatically set by the data
						resolution			= NULL,			# either NULL or an integer, specifying the temporal resolution for downsampling curves. May be needed in order to avoid excessively large data & plot files.
						xlabel				= "x", 			# e.g. 'age (years)'
						ylabel				= "y",			# e.g. 'number of lineages'
						plot_log_values		= FALSE,
						horizontal_reference= NULL,			# optional numeric, specifying a horizontal reference line to show (e.g., at value 0)
						legend_pos			= "outside",	# (string) specification of legened position, e.g. "outside". If "none", no legend will be shown
						plot_title			= "",
						data_file_comments	= "",
						save_data			= TRUE,			# (boolean)
						scatterpoints		= NULL,			# optional 2D numeric matrix of size NP x 2 (points only, X & Y) or NP x 4 (points & vertical error bars, X & Y & minY & maxY) or NP x 6 (points & vertical & horizontal error bars, X & Y & minX & maxX & minY & maxY), specifying a list of scatterpoints to add to the figure
						verbose 			= FALSE,
						verbose_prefix 		= "  "){
	curves  = curves[sapply(curves,FUN = function(l) !is.null(l))] # remove NULL elements
	Ncurves = length(curves);
	if(is.null(curve_colors)){ curve_colors = PLOT_COLOR_PALETTE[1 + (seq_len(Ncurves)-1)%%length(PLOT_COLOR_PALETTE)]; }
	else if(length(curve_colors)==1){ curve_colors = rep(curve_colors,times=Ncurves); }
	if(is.null(curve_line_types)){ curve_line_types = PLOT_LINE_TYPE_PALETTE[1 + (seq_len(Ncurves)-1)%%length(PLOT_LINE_TYPE_PALETTE)] }
	else if(length(curve_line_types)==1){ curve_line_types = rep(curve_line_types,times=Ncurves); }
	if(is.null(curve_widths)){ curve_widths = PLOT_LINE_WIDTH_PALETTE[1 + (seq_len(Ncurves)-1)%%length(PLOT_LINE_WIDTH_PALETTE)] }
	else if(length(curve_widths)==1){ curve_widths = rep(curve_widths,times=Ncurves); }
	if(is.null(show_curves)) show_curves = rep(TRUE,Ncurves)

	if(!is.null(resolution)){
		if(verbose) cat(sprintf("%sDownsampling curves to %d time points..\n",resolution))
		for(n in seq_len(Ncurves)){
			if(sum(!is.na(curves[[n]][[2]]))<2) next;
			X = curves[[n]][[1]];
			if(length(X)>resolution){
				Xnew = seq(from=X[1], to=tail(X,1), length.out=resolution);
				curves[[n]][[1]] = Xnew;
				curves[[n]][[2]] = approx(x=X, y=curves[[n]][[2]], xout=Xnew, method="linear", yleft=NaN, yright=NaN, rule = 1, f = 0, ties = mean)$y;
			}
		}
	}
	
	# save data as text file
	if(save_data){
		if(verbose) cat(sprintf("%sSaving %s to table (%s)..\n",verbose_prefix,ytype,case_tag))
		output_table=sprintf("%s.tsv",file_basepath)
		check_output_file(output_table,TRUE,TRUE,"  ")
		cat(sprintf("# %s over %s (%s)\n# %s\n#\n",ylabel,xlabel,case_tag,stringr::str_replace_all(data_file_comments,"\n","\n# ")), file=output_table, append=FALSE)
		for(n in seq_len(Ncurves)){
			cat(sprintf("# %s\n%s\t%s\n",curve_names[[n]],xtype,ytype), file=output_table, append=TRUE);
			write.table(x=cbind(curves[[n]][[1]],curves[[n]][[2]]), file=output_table, append=TRUE, sep="\t", row.names=FALSE, col.names=FALSE, quote=FALSE);
			cat(sprintf("\n\n"), file=output_table, append=TRUE);
		}
	}
		
	# omit invalid values from curves
	for(n in seq_len(Ncurves)){
		if(plot_log_values){
			valids = which(is.finite(log(curves[[n]][[2]])));
		}else{
			valids = which(is.finite(curves[[n]][[2]]));
		}
		curves[[n]][[1]] = curves[[n]][[1]][valids]
		curves[[n]][[2]] = curves[[n]][[2]][valids]
	}
	
	# determine X & Y ranges
	data_miny = NA
	data_maxy = NA
	data_minx = NA
	data_maxx = NA
	for(n in seq_len(Ncurves)){
		X = curves[[n]][[1]]
		Y = curves[[n]][[2]]
		if(!is.null(plot_minx)) Y = Y[X>=plot_minx]
		if(!is.null(plot_maxx)) Y = Y[X<=plot_maxx]
		if(plot_log_values) Y = Y[Y>0]
		if(length(Y)>0){
			data_miny = (if(is.na(data_miny)) min(Y,na.rm=TRUE) else min(data_miny,min(Y,na.rm=TRUE)));
			data_maxy = (if(is.na(data_maxy)) max(Y,na.rm=TRUE) else max(data_maxy,max(Y,na.rm=TRUE)));
			data_minx = (if(is.na(data_minx)) min(X,na.rm=TRUE) else min(data_minx,min(X,na.rm=TRUE)));
			data_maxx = (if(is.na(data_maxx)) max(X,na.rm=TRUE) else max(data_maxx,max(X,na.rm=TRUE)));
		}
	}
	plot_minx	= (if((!is.null(plot_minx)) && is.finite(plot_minx)) plot_minx else data_minx)
	plot_maxx	= (if((!is.null(plot_maxx)) && is.finite(plot_maxx)) plot_maxx else data_maxx)
	plot_miny 	= (if((!is.null(plot_miny)) && is.finite(plot_miny)) plot_miny else data_miny)
	plot_maxy 	= data_maxy + (data_maxy-plot_miny)*0.1
	if(is.na(data_miny) || is.na(data_maxy) || is.na(data_minx) || is.na(data_maxx)){
		if(verbose) cat(sprintf("%sWARNING: No valid points for plotting\n",verbose_prefix))
		return();
	}
	
	# prepare for plotting
	if(verbose) cat(sprintf("%sPlotting %s over age (%s)..\n",verbose_prefix,ytype,case_tag))
	curves 				= curves[show_curves]
	curve_names			= curve_names[show_curves]
	curve_colors		= curve_colors[show_curves]
	curve_line_types	= curve_line_types[show_curves]
	curve_widths		= curve_widths[show_curves]
	plot_file 			= sprintf("%s.pdf",file_basepath)
	Ncurves				= length(curves)
	check_output_file(plot_file,TRUE,TRUE,"  ")
	margins				= c((if(legend_pos=="none") 1 else max(1,Ncurves*0.23-DEFAULT_PLOT_HEIGHT)),1,DEFAULT_PLOT_TOP_MARGIN,(if(legend_pos=="outside") DEFAULT_PLOT_RIGHT_MARGIN_WITH_LEGEND else 0.5))
	pdf(file=plot_file, width=DEFAULT_PLOT_WIDTH+margins[2]+margins[4], height=DEFAULT_PLOT_HEIGHT+margins[1]+margins[3])
	par(mai=margins)
	
	 # initialize with empty plot
	suppressWarnings(plot(	x		= c(),
							y		= c(), 
							main 	= plot_title, 
							xlab 	= xlabel, 
							ylab 	= NA, 
							log 	= (if(plot_log_values) "y" else ""),
							yaxt	= (if(plot_log_values) "n" else NULL), 
							cex		= 1.1,
							las		= 1,
							xaxs	= "i",
							xlim	= (if(reverse_x) c(plot_maxx, plot_minx) else c(plot_minx,plot_maxx)), 
							ylim	= c(plot_miny, plot_maxy)))
	title(ylab=ylabel, line=4)
	
	# plot shadings
	if(!is.null(shadings)){
		for(sh in seq_len(length(shadings))){
			curve1 		= shadings[[sh]][[1]]
			curve2 		= shadings[[sh]][[2]]
			shade_color = shadings[[sh]][[3]]
			order1		= order(curves[[curve1]][[1]])
			order2		= order(curves[[curve2]][[1]])
			polygon(c(curves[[curve1]][[1]][order1],rev(curves[[curve2]][[1]][order2])),
					c(curves[[curve2]][[2]][order2],rev(curves[[curve1]][[2]][order1])), 
					col=shade_color,
					border=NA)
		}
	}
	
	# plot reference curve
	if(!is.null(horizontal_reference)){
		lines(	x	= c(plot_minx,plot_maxx),
				y	= c(horizontal_reference,horizontal_reference),
				type= "l",
				col	= REFERENCE_CURVE_COLOR,
				lty	= 1,
				lwd	= 1)
	}
		
	# plot curves
	for(n in seq_len(Ncurves)){
		lines(	x	= curves[[n]][[1]],
				y	= curves[[n]][[2]],
				type= "l",
				col	= curve_colors[n],
				lty	= curve_line_types[n],
				lwd	= curve_widths[n]);				
	}
	
	# plot scatterpoints
	if((!is.null(scatterpoints)) && (nrow(scatterpoints)>0)){
		points(x=scatterpoints[,1], y=scatterpoints[,2], pch=21, col="#303030", bg="#909090")
		if(ncol(scatterpoints)==4){
			# include vertical error bars
			arrows(scatterpoints[,1], scatterpoints[,3], scatterpoints[,1], scatterpoints[,4], length=0.05, angle=90, code=3, col="#303030")
		}else if(ncol(scatterpoints)==6){
			# include horizontal & vertical error bars
			arrows(scatterpoints[,3], scatterpoints[,2], scatterpoints[,4], scatterpoints[,2], length=0.05, angle=90, code=3, col="#303030") # horizontal bars
			arrows(scatterpoints[,1], scatterpoints[,5], scatterpoints[,1], scatterpoints[,6], length=0.05, angle=90, code=3, col="#303030") # vertical bars
		}
	}
	
	# add legend
	if(legend_pos!="none"){
		if(legend_pos=="outside"){
			if(reverse_x){
				legendx = plot_minx - 0.2*(plot_maxx-plot_minx)
			}else{
				legendx = plot_maxx + 0.2*(plot_maxx-plot_minx)
			}
			legendy = plot_maxy
			legend(x=legendx, y=legendy, legend = curve_names, col=curve_colors, lty=curve_line_types, lwd=curve_widths, xpd=NA);
		}else{
			legend(legend_pos, legend = curve_names, col=curve_colors, lty=curve_line_types, lwd=curve_widths);
		}
	}
	if(plot_log_values){
		# improve appearance of y-axis ticks if logarithmic
		aty = axTicks(2)
		axis(2,at=aty,labels=sapply(aty, function(x) sprintf("%g",x)), las=1)
	}
	invisible(dev.off());
}


expand_range = function(xmin,xmax,factor,logarithmic=FALSE){
	if(logarithmic){
		return(exp(expand_range(log(xmin),log(xmax),factor,logarithmic=FALSE)));
	}else{
		L = xmax - xmin;
		M = 0.5*(xmin+xmax);
		return(c(M-factor*L/2, M+factor*L/2))
	}
}


pairwise_scatterplots = function(	output_basepath,	# e.g. 'output/loglikelihoods'
									case_tag,			# character, will be included as subtitle in the plots
									scattervalues, 		# 2D numeric matrix of size NP x ND, storing NP scattered values for each of ND data types. Each scatterplot will show those values for a pair of data types, e.g. Y[,i] vs Y[,j]
									data_names,			# character vector of size ND. Can be NULL, in which case the column names of scattervalues are used as names.
									point_names,		# character vector of size NP. Can be NULL, in which case the row names of scattervalues are used as names.
									logarithmic,		# boolean vector of size ND, indicating if data types should be plotted on a log axis. Can also be a single boolean, applying to all data types
									include_diagonal,	# (boolean)
									comment_line,		# character, optional comment line to include in TSV file
									verbose,
									verbose_prefix){
	ND 	= ncol(scattervalues);
	NP	= nrow(scattervalues);
	if(is.null(data_names)) data_names = colnames(scattervalues)
	if(is.null(point_names)){
		point_names = rownames(scattervalues)
	}else{
		rownames(scattervalues) = point_names
	}
	if(length(logarithmic)==1) logarithmic  = rep(logarithmic,times=ND)
	
	# save to file
	if(verbose) cat(sprintf("%sSaving scattervalues for '%s'..\n",verbose_prefix,case_tag))
	if(endsWith(output_basepath,"/")){
		output_table=sprintf("%s%s_vs_%s.tsv",output_basepath,data_names[d1],data_names[d2])
	}else{
		output_table=sprintf("%s.tsv",output_basepath)
	}
	check_output_file(output_table,TRUE,TRUE,"  ")
	cat(sprintf("# Scatterdata for %d variables, '%s'\n# %s\n#\n# \t%s\n",ND,case_tag,comment_line,paste(data_names,collapse="\t")), file=output_table, append=FALSE)
	write.table(x=scattervalues, file=output_table, append=TRUE, sep="\t", row.names=(!is.null(point_names)), col.names=FALSE, quote=FALSE);

	# plot to PDF
	if(ND<=1) return(); # nothing to plot
	for(d1 in 1:ND){
		for(d2 in 1:ND){
			if(d1<=d2) next;
			if(verbose) cat(sprintf("%sGenerating scatterplot '%s' vs '%s', %s..\n",verbose_prefix,data_names[d1],data_names[d2],case_tag))
			valids = which(is.finite(scattervalues[,d1]) & is.finite(scattervalues[,d2]))
			if(length(valids)==0){
				if(verbose) cat(sprintf("%sWARNING: No valid points to plot, skipping scatterplot\n",verbose_prefix))
				next;
			}
			if((ND==2) && (!endsWith(output_basepath,"/"))){
				plot_file = sprintf("%s.pdf",output_basepath)
			}else{
				plot_file = sprintf("%s%s%s_vs_%s.pdf",output_basepath,(if(endsWith(output_basepath,"/")) "" else "_"),data_names[d1],data_names[d2])
			}
			check_output_file(plot_file,TRUE,TRUE,"  ")
			margins = c(0.5,1.2,DEFAULT_PLOT_TOP_MARGIN,DEFAULT_PLOT_RIGHT_MARGIN)
			pdf(file=plot_file, width=DEFAULT_PLOT_WIDTH+margins[2]+margins[4], height=DEFAULT_PLOT_HEIGHT+margins[1]+margins[3])
			X	= scattervalues[valids,d1]
			Y	= scattervalues[valids,d2]
			graphics::plot(	x = X,
							y = Y,
							type = "p",
							main = sprintf("%s vs %s\n%s",data_names[d1],data_names[d2],case_tag),
							xlab=sprintf("%s",data_names[d1]),
							ylab=sprintf("%s",data_names[d2]),
							col = "#303030",
							pch=1,
							xlim=(if(logarithmic[d1]) NULL else expand_range(min(X),max(X),1.1)),
							ylim=(if(logarithmic[d2]) NULL else expand_range(min(Y),max(Y),1.1)),
							log=paste0((if(logarithmic[d1]) "x" else ""), (if(logarithmic[d2]) "y" else "")),
							xaxt=(if(logarithmic[d1]) "n" else NULL), 
							yaxt=(if(logarithmic[d2]) "n" else NULL));
			# add diagonal
			if(include_diagonal){
				graphics::abline(a=0, b=1, col="#909090")
			}
			# improve axis tics if logarithmic
			if(logarithmic[d1]){
				atx = axTicks(1)
				axis(1,at=atx,labels=sapply(atx, function(x) sprintf("%g",x)))
			}
			if(logarithmic[d2]){
				aty = axTicks(2)
				axis(2,at=aty,labels=sapply(aty, function(y) sprintf("%g",y)), las=2)
			}
			invisible(dev.off());
		}
	}
}






plot_fitted_vs_true_model = function(	plot_dir,
										case_tag,		# e.g. 'BACTERIA_EMBL - fitted model, grid size 10'
										subtitle,
										true_model_name,
										fit_model_name,
										sim_true,		# deterministic simulation of the true model
										sim_fit,		# deterministic simulation of the fitted model
										tree_LTT,		# list containing times[] and lineages[], each of length NT
										root_age,
										time_unit,		# e.g. 'Myr'
										verbose,
										verbose_prefix){
	if(!is.null(subtitle)){ subtitle = sprintf("%s\n%s",case_tag,subtitle); }else{ subtitle = case_tag; }
	plot_curves(file_basepath		= sprintf("%s/LTT_vs_true",plot_dir),
				xtype				= 'age',
				ytype				= 'LTTs',
				case_tag			= case_tag,
				curves				= list(list(tree_LTT$ages,tree_LTT$lineages), list(sim_fit$ages,sim_fit$LTT), list(sim_true$ages,sim_true$LTT)),
				curve_names			= c("tree", "true", "fit"),
				curve_line_types	= c(1,1,2),
				curve_widths		= 1.5,
				reverse_x			= TRUE,
				plot_minx			= 0,
				plot_maxx			= root_age,
				plot_miny			= 0,
				resolution			= PLOT_DOWNSAMPLING_RESOLUTION,
				xlabel				= sprintf("age (%s)",time_unit),
				ylabel				= "lineages",
				legend_pos			= "outside",
				plot_title			= sprintf("LTT of true tree (%s)\nvs. fit model (%s)\n%s",true_model_name,fit_model_name,subtitle),
				data_file_comments	= sprintf("LTT of tree generated by true model (%s) vs deterministic LTT of fit skyline model (%s)",true_model_name,fit_model_name),
				verbose 			= FALSE,
				verbose_prefix 		= "  ")
	plot_curves(file_basepath		= sprintf("%s/PSR_vs_true",plot_dir),
				xtype				= 'age',
				ytype				= 'PSR',
				case_tag			= case_tag,
				curves				= list(list(sim_true$ages,sim_true$PSR), list(sim_fit$ages,sim_fit$PSR)),
				curve_names			= c("true", "fit"),
				curve_line_types	= c(1,2),
				curve_widths		= c(1.5,2.5),
				curve_colors 		= BLACK_CURVE_COLOR,
				reverse_x			= TRUE,
				plot_minx			= 0,
				plot_maxx			= root_age,
				plot_miny			= 0,
				resolution			= PLOT_DOWNSAMPLING_RESOLUTION,
				xlabel				= sprintf("age (%s)",time_unit),
				ylabel				= sprintf("PSR (1/%s)",time_unit),
				legend_pos			= "outside",
				plot_title			= sprintf("True PSR (%s)\nvs. fit model (%s)\n%s",true_model_name,fit_model_name,subtitle),
				data_file_comments	= sprintf("Pulled speciation rate (PSR) of true model (%s) vs fit model (%s)",true_model_name,fit_model_name),
				verbose 			= FALSE,
				verbose_prefix 		= "  ")
	plot_curves(file_basepath		= sprintf("%s/PDR_vs_true",plot_dir),
				xtype				= 'age',
				ytype				= 'PDR',
				case_tag			= case_tag,
				curves				= list(list(sim_true$ages,sim_true$PDR), list(sim_fit$ages,sim_fit$PDR)),
				curve_names			= c("true", "fit"),
				curve_line_types	= c(1,2),
				curve_widths		= c(1.5,2.5),
				curve_colors 		= BLACK_CURVE_COLOR,
				reverse_x			= TRUE,
				plot_minx			= 0,
				plot_maxx			= root_age,
				resolution			= PLOT_DOWNSAMPLING_RESOLUTION,
				xlabel				= sprintf("age (%s)",time_unit),
				ylabel				= sprintf("PDR (1/%s)",time_unit),
				legend_pos			= "outside",
				plot_title			= sprintf("True PDR (%s)\nvs. fit model (%s)\n%s",true_model_name,fit_model_name,subtitle),
				data_file_comments	= sprintf("Pulled diversification rate (PDR) of true model (%s) vs fit model (%s)",true_model_name,fit_model_name),
				verbose 			= FALSE,
				verbose_prefix 		= "  ")
	plot_curves(file_basepath 		= sprintf("%s/lambda_mu_vs_true",plot_dir),
				xtype				= "age",
				ytype	 			= "speciation/extinction rates",
				case_tag 			= case_tag,
				curves 				= list(	list(sim_true$ages,sim_true$lambda),
											list(sim_true$ages,sim_true$mu),
											list(sim_fit$ages, sim_fit$lambda),
											list(sim_fit$ages, sim_fit$mu)),
				curve_names 		= c("true lambda", "true mu", "fit lambda", "fit mu"),
				curve_colors 		= c(BLACK_CURVE_COLOR, "red", BLACK_CURVE_COLOR, "red"),
				curve_line_types	= c(1,1,2,2),
				curve_widths 		= c(1.5,2.5,1.5,2.5),
				reverse_x			= TRUE,
				plot_maxx 			= root_age,
				plot_miny			= (if(min(sim_fit$mu,na.rm=TRUE)<0) NULL else 0),
				resolution			= PLOT_DOWNSAMPLING_RESOLUTION,
				xlabel 				= sprintf("age (%s)",time_unit),
				ylabel 				= sprintf("rate (1/%s)",time_unit),
				legend_pos 			= "outside",
				plot_title			= sprintf("True lambda & mu (%s)\nvs. fit model (%s)\n%s",true_model_name,fit_model_name,subtitle),
				data_file_comments	= sprintf("Pulled lambda &  mu of true model (%s) vs fit model (%s)",true_model_name,fit_model_name),
				verbose 			= FALSE,
				verbose_prefix 		= "      ")

}




plot_fitted_model = function(	plot_dir,
								case_tag,		# e.g. 'BACTERIA_EMBL - grid size 10'
								model_name,		# e.g. 'HBD model' or 'HBDlin'
								fsimulation,
								empirical_LTT,
								root_age,
								time_unit,		# e.g. 'Myr'
								only_class,		# (boolean) only plot quantities that are constant across a congruence class
								verbose,
								verbose_prefix){
	plot_curves(file_basepath 		= sprintf("%s/LTT",plot_dir),
				xtype				= "age",
				ytype 				= "LTT",
				case_tag 			= case_tag,
				curves 	= list(	list(fsimulation$ages, fsimulation$LTT), 
								list(root_age-empirical_LTT$times, empirical_LTT$lineages)),
				curve_names 		= c("fitted dLTT", "empirical LTT"),
				curve_colors 		= c(PLOT_COLOR_FITTED, PLOT_COLOR_EMPIRICAL),
				curve_line_types	= c(PLOT_LINE_TYPE_FITTED, PLOT_LINE_TYPE_EMPIRICAL),
				curve_widths 		= c(3,1),
				reverse_x			= TRUE,
				plot_maxx 			= root_age,
				plot_miny			= 1,
				resolution			= PLOT_DOWNSAMPLING_RESOLUTION,
				xlabel 				= sprintf("age (%s)",time_unit),
				ylabel		 		= "lineages",
				legend_pos 			= "outside",
				plot_title 			= sprintf("LTT (fitted %s vs tree)\n%s",model_name,case_tag),
				data_file_comments 	= sprintf("# LTT in tree and fitted %s (%s)",model_name,case_tag),
				verbose 			= verbose,
				verbose_prefix 		= verbose_prefix)
	plot_curves(file_basepath 		= sprintf("%s/PSR",plot_dir),
				xtype				= "age",
				ytype 				= "PSR",
				case_tag 			= case_tag,
				curves 	= list(	list(fsimulation$ages, fsimulation$PSR)),
				curve_names 		= c("fitted PSR"),
				curve_colors 		= c(PLOT_COLOR_FITTED),
				curve_line_types	= c(PLOT_LINE_TYPE_FITTED),
				curve_widths 		= c(3,1),
				reverse_x			= TRUE,
				plot_maxx 			= root_age,
				plot_miny			= 0,
				resolution			= PLOT_DOWNSAMPLING_RESOLUTION,
				xlabel 				= sprintf("age (%s)",time_unit),
				ylabel		 		= "PSR",
				legend_pos 			= "outside",
				plot_title 			= sprintf("Fitted PSR\n%s",model_name,case_tag),
				data_file_comments 	= sprintf("# Fitted PSR, %s (%s)",model_name,case_tag),
				verbose 			= verbose,
				verbose_prefix 		= verbose_prefix)
	if(!only_class){
		plot_curves(file_basepath 		= sprintf("%s/lambda_mu",plot_dir),
					xtype				= "age",
					ytype	 			= "speciation/extinction rates",
					case_tag 			= case_tag,
					curves = list(	list(fsimulation$ages, fsimulation$lambda),
									list(fsimulation$ages, fsimulation$mu)),
					curve_names 		= c("fitted lambda", "fitted_mu"),
					curve_colors 		= c(PLOT_COLOR_FITTED, PLOT_COLOR_FITTED2),
					curve_line_types	= c(PLOT_LINE_TYPE_FITTED, PLOT_LINE_TYPE_FITTED2),
					curve_widths 		= c(1,3),
					reverse_x			= TRUE,
					plot_maxx 			= root_age,
					plot_miny			= (if(min(fsimulation$mu,na.rm=TRUE)<0) NULL else 0),
					resolution			= PLOT_DOWNSAMPLING_RESOLUTION,
					xlabel 				= sprintf("age (%s)",time_unit),
					ylabel 				= sprintf("rate (1/%s)",time_unit),
					legend_pos 			= "outside",
					plot_title 			= sprintf("lambda & mu of fitted %s\n%s",model_name,case_tag),
					data_file_comments 	= sprintf("# Speciation/extinction rates of fitted %s, %s",model_name,case_tag),
					verbose 			= verbose,
					verbose_prefix 		= verbose_prefix)
	}
}



# calculate equal-tailed credible intervals of a collection of curves
calculate_equal_tailed_CIs = function(curves){ 	# 2D numeric matrix of size NB x NG, listing the values of NB curves on NG grid points
	NB 	= nrow(curves)
	NG	= ncol(curves)
	CI50lower 	= numeric(NG)
	CI50upper 	= numeric(NG)
	CI95lower 	= numeric(NG)
	CI95upper 	= numeric(NG)
	medians 	= numeric(NG)
	for(g in seq_len(NG)){
		quantiles = quantile(curves[,g], probs=c(0.25, 0.75, 0.025, 0.975, 0.5), na.rm=TRUE)
		CI50lower[g] = quantiles[1]
		CI50upper[g] = quantiles[2]
		CI95lower[g] = quantiles[3]
		CI95upper[g] = quantiles[4]
		medians[g] 	 = quantiles[5]	
	}
	return(list(medians 	= medians,
				CI50lower 	= CI50lower,
				CI50upper 	= CI50upper,
				CI95lower	= CI95lower,
				CI95upper	= CI95upper))
}



# assess the adequacy of a (presumably fitted) model to explaining a given (presumably real) tree
# model should be a list containing at least the following:
#	ages : numeric vector of size NG
#	PSR : numeric vector of size NG
assess_model_adequacy = function(	tree, 
									model,
									tree_name,
									model_name,
									Nbootstraps,
									report_file = "",
									LTT_plot_basepath = "",
									Nthreads = 1){
	adequacy = castor::model_adequacy_hbd(tree, models=model, splines_degree=1, extrapolate=TRUE, Nbootstraps = Nbootstraps, Nthreads=Nthreads)
	if(!adequacy$success) return(list(success=FALSE, error=sprintf("Failed to test model adequacy: %s",adequacy$error)))

	# save report to a text file
	if(report_file!=""){
		dir.create(dirname(report_file), showWarnings = FALSE, recursive=TRUE)
		sink(file=report_file); print(adequacy); sink();
	}
	
	# plot tree & bootstrap LTTs
	if(LTT_plot_basepath!=""){
		plot_curves(file_basepath		= LTT_plot_basepath,
					xtype				= 'age',
					ytype				= "LTT",
					case_tag			= sprintf("%s - %s",tree_name,model_name),
					curves				= list(		list(adequacy$LTT_ages,adequacy$bootstrap_LTT_CI$CI95lower), 
													list(adequacy$LTT_ages,adequacy$bootstrap_LTT_CI$CI95upper), 
													list(adequacy$LTT_ages,adequacy$bootstrap_LTT_CI$CI50lower), 
													list(adequacy$LTT_ages,adequacy$bootstrap_LTT_CI$CI50upper), 
													list(adequacy$LTT_ages,adequacy$bootstrap_LTT_CI$medians),
													list(adequacy$LTT_ages,adequacy$tree_LTT)),
					curve_names			= c("CI95lower", "CI95upper", "CI50lower", "CI50upper","median","tree"),
					curve_colors		= c(CI95_CURVE_COLOR, CI95_CURVE_COLOR, CI50_CURVE_COLOR, CI50_CURVE_COLOR, BLACK_CURVE_COLOR, BLUE_CURVE_COLOR),
					curve_line_types	= 1,
					curve_widths		= c(0.4,0.4,0.4,0.4,1.4,1.4),
					shadings			= list(list(1,2,CI95_SHADE_COLOR), list(3,4,CI50_SHADE_COLOR)),
					reverse_x			= TRUE,
					plot_minx			= 0,
					plot_maxx			= root_age,
					plot_miny			= 0,
					resolution			= NULL,
					xlabel				= "age",
					ylabel				= "LTT",
					legend_pos			= "outside",
					plot_title			= sprintf("LTT equal-tailed credible intervals\ntree %s, model %s",tree_name,model_name),
					data_file_comments	= sprintf("Equal-tailed credible intervals of LTT, tree %s vs model %s",tree_name,model_name),
					verbose 			= FALSE)
	}				
	return(adequacy)
}




# Fit BD-ELC models to a tree and choose the best fit model according to AIC
fit_ELC_models = function(	tree, 
							guess_lambda	= NULL, 	# if NULL, will be automatically chosen,
							guess_mu		= NULL, 	# if NULL, will be automatically chosen
							param_guesses	= NULL,		# named list, whose elements are named lists, specifying optional parameter guesses for all (or some) model types. For example, param_guesses[["lambda_const_mu_exp"]] is a named list specifying initial parameter values for the model "lambda_const_mu_exp". param_guesses[] need not list all model types.
							rho				= 1,
							oldest_age 		= NULL,		# if NULL, will be set to the root age
							Ntrials			= 1,
							negative_mu		= FALSE,	# whether to allow negative mu
							fixed_lambda	= NULL,		# optional function handle, mapping ages to the fixed lambda
							only_model_types= NULL,		# optional character vector, specifying which models to consider. If NULL, all ELC models are considered.
							verbose			= FALSE,
							verbose_prefix	= ""){
	root_age 		= get_tree_span(tree)$max_distance
	Ntips			= length(tree$tip.label)
	lambda_types 	= (if(is.null(fixed_lambda)) c("const", "exp", "exp_plus_const", "linear") else c("fixed"))
	mu_types		= c("const", "exp", "exp_plus_const", "linear")
	NL				= length(lambda_types)
	NM				= length(mu_types)
	if(is.null(oldest_age)) oldest_age = root_age
	fit_age_grid	= seq(from=0,to=oldest_age,length.out=1000)
	
	# guesstimate lambda & mu if not provided, by fitting a const-rates model
	if(!is.null(fixed_lambda)) guess_lambda = fixed_lambda(0)
	if(is.null(guess_lambda) || is.null(guess_mu)){
		if(verbose) cat2(sprintf("%sFitting BDconst for guesstimating lambda and/or mu..\n",verbose_prefix))
		fit = fit_hbd_model_on_grid(tree, 
									oldest_age			= oldest_age,
									age_grid			= NULL,
									fixed_lambda		= (if(is.null(guess_lambda)) NULL else guess_lambda),
									fixed_mu			= (if(is.null(guess_mu)) NULL else guess_mu),
									fixed_rho			= rho,
									condition			= "auto",
									relative_dt			= INTEGRATION_RELATIVE_DT,
									Ntrials				= Ntrials,
									Nthreads			= NUMBER_OF_PARALLEL_THREADS,
									max_model_runtime	= fitting_Ntips2max_model_runtime(Ntips),
									fit_control			= list(eval.max=FITTING_NEVALUATIONS, iter.max=FITTING_NITERATIONS, rel.tol=FITTING_REL_TOLERANCE, step.min=FITTING_STEP_MIN))
		if(!fit$success) return(list(success=FALSE, error="Could not fit BDconst for guesstimating lambda & mu"))
		if(is.null(guess_lambda)) guess_lambda = fit$fitted_lambda
		if(is.null(guess_mu)) guess_mu = fit$fitted_mu
		if(guess_mu==0) guess_mu = guess_lambda/2
		if(verbose) cat2(sprintf("%s --> Guess lambda=%g, mu=%g, rho=%g\n",verbose_prefix,guess_lambda,guess_mu,rho))
	}
	
	model_types		= character(NL*NM)
	loglikelihoods	= rep(NA,NL*NM)
	AICs			= rep(NA,NL*NM)
	model_fits		= vector(mode="list", NL*NM)
	lambda0s		= rep(NA,NL*NM)
	mu0s			= rep(NA,NL*NM)
	lambda_funs		= vector(mode="list", NL*NM)
	minmus			= rep(NA,NL*NM)
	mu_funs			= vector(mode="list", NL*NM)
	fitted_params	= vector(mode="list", NL*NM)
	for(lt in 1:NL){
		for(mt in 1:NM){
			mo = NM*(lt-1) + mt
			model_types[mo] = sprintf("lambda_%s_mu_%s",lambda_types[lt],mu_types[mt])
			if((!is.null(only_model_types)) && (!(model_types[mo] %in% only_model_types))) next; # skip this model, as requested
			if(verbose) cat2(sprintf("%sFitting model type '%s'..\n",verbose_prefix,model_types[mo]))
			# prepare model structure
			# split model into a lambda- and mu-part, then combine at the end
			if(lambda_types[lt]=="const"){
				lambda_funs[[mo]] 	= function(ages,params){ pmax(0,rep(params[["lambda0"]],times=length(ages))) }
				guess_lambda_params = c(lambda0=guess_lambda)
				min_lambda_params	= c(lambda0=0)
				max_lambda_params	= c(lambda0=100*guess_lambda)
				lambda_param_scales	= c(lambda0=abs(guess_lambda))
			}else if(lambda_types[lt]=="linear"){
				lambda_funs[[mo]] 	= function(ages,params){ pmax(0,params[["lambda0"]] * (1-ages/root_age) + params[["lambda1"]] * (ages/root_age)) }
				guess_lambda_params = c(lambda0=guess_lambda, lambda1=guess_lambda)
				min_lambda_params	= c(lambda0=0, lambda1=0)
				max_lambda_params	= c(lambda0=100*guess_lambda, lambda1=100*guess_lambda)
				lambda_param_scales	= c(lambda0=abs(guess_lambda), lambda1=abs(guess_lambda))
			}else if(lambda_types[lt]=="exp"){
				lambda_funs[[mo]] 	= function(ages,params){ pmax(0,params[["lambda0"]] * exp(params[["lambda_slope"]]*ages/root_age)) }
				guess_lambda_params = c(lambda0=guess_lambda, lambda_slope=0)
				min_lambda_params	= c(lambda0=0, lambda_slope=-5)
				max_lambda_params	= c(lambda0=100*guess_lambda, lambda_slope=+5)
				lambda_param_scales	= c(lambda0=abs(guess_lambda), lambda_slope=1)
			}else if(lambda_types[lt]=="exp_plus_const"){
				lambda_funs[[mo]] 	= function(ages,params){ pmax(0,params[["lambda1"]] + params[["lambda0"]] * exp(params[["lambda_slope"]]*ages/root_age)) }
				guess_lambda_params = c(lambda0=guess_lambda/2, lambda1=guess_lambda/2, lambda_slope=0)
				min_lambda_params	= c(lambda0=0, lambda1=0, lambda_slope=-5)
				max_lambda_params	= c(lambda0=100*guess_lambda, lambda1=100*guess_lambda, lambda_slope=+5)
				lambda_param_scales	= c(lambda0=abs(guess_lambda), lambda1=abs(guess_lambda), lambda_slope=1)
			}else if(lambda_types[lt]=="exp_plus_linear"){
				lambda_funs[[mo]] 	= function(ages,params){ pmax(0,params[["lambda1"]] * (1-ages/root_age) + params[["lambda2"]] * (ages/root_age) + params[["lambda0"]] * exp(params[["lambda_slope"]]*ages/root_age)) }
				guess_lambda_params = c(lambda0=guess_lambda/2, lambda1=guess_lambda/2, lambda2=guess_lambda/2, lambda_slope=0)
				min_lambda_params	= c(lambda0=0, lambda1=0, lambda2=0, lambda_slope=-5)
				max_lambda_params	= c(lambda0=100*guess_lambda, lambda1=100*guess_lambda, lambda2=100*guess_lambda, lambda_slope=+5)
				lambda_param_scales	= c(lambda0=abs(guess_lambda), lambda1=abs(guess_lambda), lambda2=abs(guess_lambda), lambda_slope=1)
			}else if(lambda_types[lt]=="fixed"){
				lambda_funs[[mo]] 	= function(ages,params){ fixed_lambda(ages) }
				guess_lambda_params = c()
				min_lambda_params	= c()
				max_lambda_params	= c()
				lambda_param_scales	= c()
			}
			if(mu_types[mt]=="const"){
				mu_funs[[mo]] 	= function(ages,params){ pmax((if(negative_mu) -Inf else 0), rep(params[["mu0"]],times=length(ages))) }
				guess_mu_params = c(mu0=guess_mu)
				min_mu_params	= c(mu0=(if(negative_mu) -Inf else 0))
				max_mu_params	= c(mu0=100*guess_mu)
				mu_param_scales	= c(mu0=abs(guess_mu))
			}else if(mu_types[mt]=="linear"){
				mu_funs[[mo]] 	= function(ages,params){ pmax((if(negative_mu) -Inf else 0),params[["mu0"]] * (1-ages/root_age) + params[["mu1"]] * (ages/root_age)) }
				guess_mu_params = c(mu0=guess_mu, mu1=guess_mu)
				min_mu_params	= c(mu0=(if(negative_mu) -Inf else 0), mu1=(if(negative_mu) -Inf else 0))
				max_mu_params	= c(mu0=100*guess_mu, mu1=100*guess_mu)
				mu_param_scales	= c(mu0=abs(guess_mu), mu1=abs(guess_mu))
			}else if(mu_types[mt]=="exp"){
				mu_funs[[mo]] 	= function(ages,params){ pmax((if(negative_mu) -Inf else 0), params[["mu0"]] * exp(params[["mu_slope"]]*ages/root_age)) }
				guess_mu_params = c(mu0=guess_mu, mu_slope=0)
				min_mu_params	= c(mu0=(if(negative_mu) -Inf else 0), mu_slope=-5)
				max_mu_params	= c(mu0=100*guess_mu, mu_slope=+5)
				mu_param_scales	= c(mu0=abs(guess_mu), mu_slope=1)
			}else if(mu_types[mt]=="exp_plus_const"){
				mu_funs[[mo]] 	= function(ages,params){ pmax((if(negative_mu) -Inf else 0), params[["mu1"]] + params[["mu0"]] * exp(params[["mu_slope"]]*ages/root_age)) }
				guess_mu_params = c(mu0=guess_mu/2, mu1=guess_mu/2, mu_slope=0)
				min_mu_params	= c(mu0=(if(negative_mu) -Inf else 0), mu1=(if(negative_mu) -Inf else 0), mu_slope=-5)
				max_mu_params	= c(mu0=100*guess_mu, mu1=100*guess_mu, mu_slope=+5)
				mu_param_scales	= c(mu0=abs(guess_mu), mu1=abs(guess_mu), mu_slope=1)
			}else if(mu_types[mt]=="exp_plus_linear"){
				mu_funs[[mo]] 	= function(ages,params){ pmax((if(negative_mu) -Inf else 0),params[["mu1"]] * (1-ages/root_age) + params[["mu2"]] * (ages/root_age) + params[["mu0"]] * exp(params[["mu_slope"]]*ages/root_age)) }
				guess_mu_params = c(mu0=guess_mu/2, mu1=guess_mu/2, mu2=guess_mu/2, mu_slope=0)
				min_mu_params	= c(mu0=(if(negative_mu) -Inf else 0), mu1=(if(negative_mu) -Inf else 0), mu2=(if(negative_mu) -Inf else 0), mu_slope=-5)
				max_mu_params	= c(mu0=100*guess_mu, mu1=100*guess_mu, mu2=100*guess_mu, mu_slope=+5)
				mu_param_scales	= c(mu0=abs(guess_mu), mu1=abs(guess_mu), mu2=abs(guess_mu), mu_slope=1)
			}else if(mu_types[mt]=="zero"){
				mu_funs[[mo]] 	= function(ages,params){ rep(0, times=length(ages)) }
				guess_mu_params = c()
				min_mu_params	= c()
				max_mu_params	= c()
				mu_param_scales	= c()
			}
			if(is.null(param_guesses) || is.null(param_guesses[[model_types[mo]]])){
				param_guess = c(guess_lambda_params,guess_mu_params)
			}else{
				param_guess = param_guesses[[model_types[mo]]]
			}
			param_min	= c(min_lambda_params, min_mu_params)
			param_max	= c(max_lambda_params, max_mu_params)
			param_scale	= c(lambda_param_scales, mu_param_scales)
			
			# fit parametric model
			fit = fit_hbd_model_parametric(	tree, 
											param_values		= setNames(rep(NA, length(param_guess)), names(param_guess)),
											param_guess			= param_guess,
											param_min			= param_min,
											param_max			= param_max,
											param_scale			= param_scale,
											oldest_age			= oldest_age,
											lambda				= lambda_funs[[mo]],
											mu					= mu_funs[[mo]],
											rho0				= rho,
											age_grid			= fit_age_grid,
											condition			= "auto",	# one of "crown" or "stem" or "none" or "auto", specifying whether to condition the likelihood on the survival of the stem group or the crown group. It is recommended to use "stem" when oldest_age>root_age, and "crown" when oldest_age==root_age. This argument is similar to the "cond" argument in the R function RPANDA::likelihood_bd. Note that "crown" really only makes sense when oldest_age==root_age.
											relative_dt			= INTEGRATION_RELATIVE_DT,
											Ntrials				= Ntrials,
											max_start_attempts	= 100,
											Nthreads			= NUMBER_OF_PARALLEL_THREADS,
											max_model_runtime	= fitting_Ntips2max_model_runtime(Ntips),
											fit_algorithm		= "nlminb",
											fit_control			= list(eval.max=FITTING_NEVALUATIONS, iter.max=FITTING_NITERATIONS, rel.tol=FITTING_REL_TOLERANCE, step.min=FITTING_STEP_MIN),
											diagnostics			= FALSE,
											verbose				= FALSE,
											verbose_prefix		= paste0(verbose_prefix,"  "))
			if(!fit$success){
				if(verbose) cat2(sprintf("%s  WARNING: Could not fit model type '%s': %s\n",verbose_prefix,model_types[mo],fit$error))
			}else{
				loglikelihoods[mo]	= fit$loglikelihood
				AICs[mo] 			= fit$AIC
				model_fits[[mo]]	= fit
				lambda0s[mo]		= lambda_funs[[mo]](0,fit$param_fitted)
				mu0s[mo]			= mu_funs[[mo]](0,fit$param_fitted)
				minmus[mo]			= min(mu_funs[[mo]](fit_age_grid,fit$param_fitted))
				fitted_params[[mo]]	= fit$param_fitted
				if(verbose) cat2(sprintf("%s  --> AIC=%.10g, mu0=%g\n",verbose_prefix,fit$AIC,mu0s[mo]))
			}
		}
	}
	if(all(is.na(AICs))) return(list(success=FALSE, error="Fitting failed for all model types"))
	best_model 	= which.min(AICs)
	best_fit	= model_fits[[best_model]]
	if(verbose) cat2(sprintf("%sNote: Best model '%s' has AIC=%.10g, mu0=%g, min_mu=%g\n",verbose_prefix,model_types[best_model],best_fit$AIC,mu0s[best_model],minmus[best_model]))
	
	# simulate the best model
	if(verbose) cat2(sprintf("%sSimulating the best model (%s)..\n",verbose_prefix,model_types[best_model]))
	fsim = simulate_deterministic_hbd(	LTT0			= Ntips,
										oldest_age		= oldest_age,
										rho				= rho,
										age_grid		= fit_age_grid,
										lambda			= lambda_funs[[best_model]](fit_age_grid,best_fit$param_fitted),
										mu				= mu_funs[[best_model]](fit_age_grid,best_fit$param_fitted),
										relative_dt		= INTEGRATION_RELATIVE_DT,
										allow_unreal	= negative_mu)
	if(!fsim$success) cat2(sprintf("%s  WARNING: Simulation failed: %s\n",verbose_prefix,fsim$error))	

	return(list(success			= TRUE,
				model_types		= model_types,
				loglikelihoods	= loglikelihoods,
				AICs			= AICs,
				model_fits		= model_fits,
				lambda0s		= lambda0s,
				mu0s			= mu0s,
				minmus			= minmus,
				fitted_params	= setNames(fitted_params,model_types),
				best_model		= best_model,
				best_fit_sim	= fsim))
}



###############################								
# PREPARATIONS

options(expressions=100000) # increase recursion depth
options(max.print=1000000) # increase print count

# print warnings as they occur
options(warn=1)

cat(sprintf("Loading %d required packages (%s)..\n",length(REQUIRED_PACKAGES),paste(REQUIRED_PACKAGES,collapse=", ")));
for(p in 1:length(REQUIRED_PACKAGES)){
	cat(sprintf("  Loading %s..\n",REQUIRED_PACKAGES[p]))
	library(REQUIRED_PACKAGES[p], quietly=TRUE, character.only=TRUE)
}

# seed random number generators
set.seed(NULL)
random_seed = sample.int(n=1000000,size=1)
set.seed(random_seed)
cat(sprintf("Note: Seeding random generator at %d\n",random_seed));


# prepare output dir
output_dir = get_non_existent_dir("output", "run_", 3);
dir.create(output_dir, showWarnings = FALSE, recursive=TRUE);
cat(sprintf("All output will be written to '%s'..\n",output_dir))


# save backup of this script
initial.options = commandArgs(trailingOnly = FALSE);
script_path <- sub("--file=", "", initial.options[grep("--file=", initial.options)]);
OK = file.copy(script_path, to = file.path(output_dir,"workflow_backup.R"));

# prepare log file
logfile = sprintf("%s/log.txt",output_dir);
cat2 = function(message, file=logfile, append=TRUE){
	cat(message);
	cat(message, file=file, append=append);
}


#############################################################
# Simulating trees under variable-rate BD models (BDvar)
# and then fitting the BDELC & BDELCNeg models
# Speciation and extinction rates of of the simulated models are chosen randomly according to stochastic processes

if(INCLUDE_BDELC_FITTING_TO_SIMULATIONS){
	cat2(sprintf("Simulating trees and fitting birth-death models..\n"))
	sim_dir = sprintf("%s/simulations_BDvar",output_dir)
	sim_results	= data.frame(	simulation					= c(1:SIM_FITTING_NSIMULATIONS),
								root_age 					= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								lambda0						= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								mu0							= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								true_minmu					= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								rho							= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELC_type		 			= character(SIM_FITTING_NSIMULATIONS),
								BDELC_loglikelihood 		= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELC_AIC					= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELC_lambda				= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELC_mu					= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELC_minmu					= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELC_epsilon				= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELC_relative_error		= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELC_PnodeKS				= rep(NA, times=SIM_FITTING_NSIMULATIONS), # P-value of Kolmogorov-Smirnov test for node ages
								BDELC_fraction_LTT_in_CI95	= rep(NA, times=SIM_FITTING_NSIMULATIONS), # fraction of the tree's LTT contained in the 95-CI of the model bootstraps
								BDELC_dLTT_R2				= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCNeg_type		 		= character(SIM_FITTING_NSIMULATIONS),
								BDELCNeg_loglikelihood 		= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCNeg_AIC				= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCNeg_lambda				= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCNeg_mu					= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCNeg_minmu				= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCNeg_epsilon			= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCNeg_relative_error		= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCNeg_dLTT_R2			= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCNeg_CTL_mu				= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCTL_type				= character(SIM_FITTING_NSIMULATIONS),
								BDELCTL_loglikelihood		= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCTL_AIC					= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCTL_mu					= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCTL_PnodeKS				= rep(NA, times=SIM_FITTING_NSIMULATIONS),
								BDELCTL_relative_error		= rep(NA, times=SIM_FITTING_NSIMULATIONS))

	n = 1 # simulation counter
	while(n<=SIM_FITTING_NSIMULATIONS){
		cat2(sprintf("  Simulation #%d..\n",n))
		this_sim_dir = sprintf("%s/simulation_%d",sim_dir,n)
		dir.create(this_sim_dir, showWarnings = FALSE, recursive=TRUE)
		tree_name = sprintf("BDvar_%d",n)
		
		# reset the random number generator and keep track of the specific seed used, to allow reproducibility of simulations when troubleshooting
		tree_seed = sample.int(n=1000000,size=1)
		set.seed(tree_seed)
		cat(sprintf("    Note: Seeding random generator at %d\n",tree_seed))
		cat2(sprintf("    Generating random tree #%d..\n",n))

		# simulate random lambda & mu time series
		rho 			= exp(runif(min=log(0.01), max=log(1), n=1)) # choose sampling fraction for this simulation
		series_ages 	= seq(from=0,to=100,by=1)
		Nsim_attempts 	= 0
		while(TRUE){
			Nsim_attempts = Nsim_attempts + 1
			lambda_stationary_mean = 1
			series_lambda = castor:::generate_OU_time_series(	times 			= series_ages,
																start_value		= lambda_stationary_mean,
																stationary_mean = lambda_stationary_mean,
																stationary_std	= 0.5*lambda_stationary_mean,
																decay_rate		= 0.1,
																constrain_min	= 0.1)$values
			mu_stationary_mean = runif(min=0.1, max=0.8, n=1) * lambda_stationary_mean
			series_mu = castor:::generate_OU_time_series(	times 			= series_ages,
															start_value		= mu_stationary_mean,
															stationary_mean = mu_stationary_mean,
															stationary_std	= 0.5*mu_stationary_mean,
															decay_rate		= 0.1,
															constrain_min	= 0.1)$values
			if(mean(series_lambda-series_mu)<=0) next # avoid scenarios where the mean net diversification rate is negative
			# generate random tree based on the specific lambda & mu
			tree_gen = generate_tree_hbd_reverse(	Ntips				= SIM_BDVAR_TREE_SIZE,
													age_grid			= series_ages,
													lambda				= series_lambda,
													mu					= series_mu,
													rho					= rho,
													force_max_age		= tail(series_ages,1),
													splines_degree		= 1)
			if(!tree_gen$success) next
			tree	 = tree_gen$trees[[1]]
			root_age = get_tree_span(tree)$max_distance
			Ntips	 = length(tree$tip.label)
			# simulate deterministic model
			sim_true = simulate_deterministic_hbd(	Ntips,
													oldest_age		= root_age,
													rho				= rho,
													age_grid		= series_ages,	
													lambda			= series_lambda,
													mu				= series_mu,
													relative_dt		= INTEGRATION_RELATIVE_DT,
													allow_unreal	= FALSE)
			if(!sim_true$success) next
			# all seems OK with this simulation
			break
		}
		sim_results$lambda0[n] 		= series_lambda[1]
		sim_results$mu0[n] 			= series_mu[1]
		sim_results$true_minmu[n] 	= min(series_mu)
		sim_results$rho[n] 			= rho

		cat2(sprintf("      Note: Sim attempts %d. Tree has %d tips, spans %g, mu0=%g, true_minmu=%g\n",Nsim_attempts,Ntips,root_age,series_mu[1],sim_results$true_minmu[n]))
		if(SIM_NOISY){
			castor::write_tree(tree, file=sprintf("%s/tree_noisefree.tre",this_sim_dir))
			cat2(sprintf("    Calculating and plotting LTT of noise-free tree..\n"))
			LTT_noisefree = castor::count_lineages_through_time(tree,Ntimes=10*log2(Ntips));
			LTT_noisefree$ages = root_age-LTT_noisefree$times
			plot_curves(file_basepath 		= sprintf("%s/LTT_noisefree",this_sim_dir),
						xtype				= "age",
						ytype	 			= "LTTs",
						case_tag 			= sprintf("simulation %s, noise-free",n),
						curves 				= list(list(LTT_noisefree$ages, LTT_noisefree$lineages), list(sim_true$ages, sim_true$LTT)),
						curve_names 		= c("tree LTT", "true dLTT"),
						curve_line_types	= 1,
						curve_widths 		= 2,
						reverse_x			= TRUE,
						plot_maxx 			= root_age,
						plot_miny			= 1,
						resolution			= PLOT_DOWNSAMPLING_RESOLUTION,
						xlabel 				= sprintf("age (%s)",SIM_TIME_UNIT),
						ylabel 				= "lineages",
						legend_pos 			= "outside",
						plot_title 			= sprintf("LTT of noise-free simulated tree #%d",n),
						data_file_comments 	= sprintf("LTT of noise-free simulated tree #%d",n),
						verbose 			= FALSE,
						verbose_prefix 		= "      ")
			
			cat2(sprintf("    Adding noise to simulated tree..\n"))
			tree = add_noise_to_tree(tree, multifurcation_fraction=NOISY_SIM_MULTIFURCATION_FRACTION, max_relative_age_error=NOISY_SIM_MAX_RELATIVE_AGE_ERROR)
			castor::write_tree(tree, file=sprintf("%s/tree.tre",this_sim_dir))
			root_age = get_tree_span(tree)$max_distance
		}else{
			castor::write_tree(tree, file=sprintf("%s/tree.tre",this_sim_dir))
		}
		fit_oldest_age = SIM_FITTING_AGE_FRACTION*root_age
		sim_results$root_age[n] = root_age
	
		cat2(sprintf("    Plotting lambda & mu for simulated tree #%d..\n",n))
		plot_curves(file_basepath		= sprintf("%s/lambda_mu",this_sim_dir),
					xtype				= "age",
					ytype				= "lambda & mu",
					case_tag			= sprintf("simulation %d",n),
					curves				= list(list(series_ages,series_lambda), list(series_ages,series_mu)),
					curve_names			= c("lambda", "mu"),
					curve_line_types	= c(1,2),
					curve_widths		= c(2,2),
					reverse_x			= TRUE,
					plot_maxx			= root_age,
					plot_miny			= 0,
					resolution			= PLOT_DOWNSAMPLING_RESOLUTION,
					xlabel 				= sprintf("age (%s)",SIM_TIME_UNIT),
					ylabel				= sprintf("rate (1/%s)",SIM_TIME_UNIT),
					legend_pos			= "outside",
					plot_title			= sprintf("lambda & mu of simulated tree #%d",n),
					data_file_comments	= sprintf("lambda & mu of simulated tree #%d",n),
					verbose 			= FALSE,
					verbose_prefix 		= "      ")
						
		# calculate tree LTT	
		cat2(sprintf("    Calculating and plotting tree LTT..\n"))
		LTT = castor::count_lineages_through_time(tree,Ntimes=10*log2(Ntips));
		LTT$ages = root_age-LTT$times
		plot_curves(file_basepath 		= sprintf("%s/LTT",this_sim_dir),
					xtype				= "age",
					ytype	 			= "LTTs",
					case_tag 			= sprintf("simulation %s",n),
					curves 				= list(list(LTT$ages, LTT$lineages), list(sim_true$ages, sim_true$LTT)),
					curve_names 		= c("tree LTT", "true dLTT"),
					curve_line_types	= 1,
					curve_widths 		= 2,
					reverse_x			= TRUE,
					plot_maxx 			= root_age,
					plot_miny			= 1,
					resolution			= PLOT_DOWNSAMPLING_RESOLUTION,
					xlabel 				= sprintf("age (%s)",SIM_TIME_UNIT),
					ylabel 				= "lineages",
					legend_pos 			= "outside",
					plot_title 			= sprintf("LTT of simulated tree #%d",n),
					data_file_comments 	= sprintf("LTT of simulated tree #%d",n),
					verbose 			= FALSE,
					verbose_prefix 		= "      ")


		# fit ELC models, not allowing for negative mu
		best_BDELC_model_type = NA
		fit_dir = sprintf("%s/fitted_BDELC",this_sim_dir)
		cat2(sprintf("    Fitting ELC models to simulated tree #%d..\n",n))
		fits = fit_ELC_models(	tree			= tree, 
								rho				= rho,
								oldest_age		= fit_oldest_age,
								Ntrials			= FITTING_NTRIALS,
								negative_mu		= FALSE,
								verbose			= TRUE,
								verbose_prefix	= "      ")
		if(!fits$success){
			cat2(sprintf("      ERROR: Could not fit ELC models: %s\n",fits$error))
			if(SIM_FITTING_REPEAT_FAILED_TREES){
				cat2(sprintf("             Repeating entire simulation %d and fitting\n",n))
				unlink(this_sim_dir, recursive=TRUE)
				next
			}
		}else{
			dir.create(fit_dir, showWarnings = FALSE, recursive=TRUE);
			sink(file=sprintf("%s/fit_results.txt",fit_dir)); print(fits$model_fits[[fits$best_model]]); sink(); # save fit results to text file
			cat2(sprintf("    Plotting fitted ELC model for simulated tree #%d..\n",n));
			plot_fitted_model(	plot_dir		= fit_dir,
								case_tag		= sprintf('tree %d',n),
								model_name		= sprintf('HBDELC - %s',fits$model_types[fits$best_model]),
								fsimulation		= fits$best_fit_sim,
								empirical_LTT	= LTT,
								root_age		= root_age,
								time_unit		= SIM_TIME_UNIT,
								only_class		= FALSE,
								verbose			= FALSE,
								verbose_prefix	= "      ")
			plot_fitted_vs_true_model(	plot_dir		= fit_dir,
										case_tag		= sprintf('tree %d',n),
										subtitle		= "Fitted BDELC",
										true_model_name	= sprintf("sim %d",n),
										fit_model_name	= "BDELC",
										sim_true		= sim_true,
										sim_fit			= fits$best_fit_sim,
										tree_LTT		= LTT,
										root_age		= root_age,
										time_unit		= SIM_TIME_UNIT,
										verbose			= FALSE,
										verbose_prefix	= "      ")
			# assess adequacy of fitted model
			cat2(sprintf("    Assessing adequacy of fitted ELC model..\n"))
			adequacy = assess_model_adequacy(	tree 				= tree, 
												model				= fits$best_fit_sim, 
												tree_name 			= tree_name, 
												model_name			= "BDELC_fit", 
												Nbootstraps			= MODEL_ADEQUACY_NBOOTSTRAPS, 
												report_file			= sprintf("%s/fit_model_adequacy.txt",fit_dir), 
												LTT_plot_basepath	= sprintf("%s/LTT_tree_vs_fitted_model",fit_dir),
												Nthreads			= NUMBER_OF_PARALLEL_THREADS)
			dLTT_R2 = get_R2(xtrue=sim_true$ages, sim_true$LTT, xfit=fits$best_fit_sim$ages, yfit=fits$best_fit_sim$LTT, minx=0, maxx=root_age) # compare dLTT of fitted vs true model
			cat2(sprintf("     --> PnodeKS=%g, fraction LTT in CI95=%g, dLTT_R2=%g\n",adequacy$PnodeKS,adequacy$fraction_LTT_in_CI95,dLTT_R2))
			# keep record of main EC fit results
			best_BDELC_model_type						= fits$model_types[fits$best_model]
			sim_results$BDELC_type[n]					= fits$model_types[fits$best_model]
			sim_results$BDELC_loglikelihood[n]			= fits$loglikelihoods[fits$best_model]
			sim_results$BDELC_AIC[n]					= fits$AICs[fits$best_model]
			sim_results$BDELC_lambda[n]					= fits$lambda0s[fits$best_model]
			sim_results$BDELC_mu[n]						= fits$mu0s[fits$best_model]
			sim_results$BDELC_minmu[n]					= fits$minmus[fits$best_model]
			sim_results$BDELC_epsilon[n]				= fits$mu0s[fits$best_model]/fits$lambda0s[fits$best_model]
			sim_results$BDELC_PnodeKS[n]				= adequacy$PnodeKS
			sim_results$BDELC_fraction_LTT_in_CI95[n]	= adequacy$fraction_LTT_in_CI95
			sim_results$BDELC_relative_error[n]			= (fits$mu0s[fits$best_model]-series_mu[1])/series_mu[1]
			sim_results$BDELC_dLTT_R2[n]				= dLTT_R2

			# fit ELC models, allowing for negative mu
			cat2(sprintf("    Fitting ELC models (allowing for negative mu) to simulated tree #%d..\n",n))
			fit_dir = sprintf("%s/fitted_BDELCNeg",this_sim_dir)
			fits = fit_ELC_models(	tree			= tree, 
									rho				= rho,
									oldest_age		= fit_oldest_age,
									Ntrials			= FITTING_NTRIALS,
									negative_mu		= TRUE,
									only_model_types= (if(SIM_FITTING_BDELCNEG_USE_BDELC_MODEL && (!is.na(best_BDELC_model_type))) stringr::str_replace_all(best_BDELC_model_type,"mu_zero","mu_const") else NULL),
									param_guesses	= (if(SIM_FITTING_BDELCNEG_USE_BDELC_AS_GUESS && fits$success) fits$fitted_params else NULL),
									verbose			= TRUE,
									verbose_prefix	= "      ")
			if(!fits$success){
				cat2(sprintf("      ERROR: Could not fit ELC models: %s\n",fits$error))
				if(SIM_FITTING_REPEAT_FAILED_TREES){
					cat2(sprintf("             Repeating entire simulation %d and fitting\n",n))
					unlink(this_sim_dir, recursive=TRUE)
					next
				}
			}else{
				dir.create(fit_dir, showWarnings = FALSE, recursive=TRUE);
				sink(file=sprintf("%s/fit_results.txt",fit_dir)); print(fits$model_fits[[fits$best_model]]); sink(); # save fit results to text file
				cat2(sprintf("    Plotting fitted ELC model for simulated tree #%d..\n",n));
				plot_fitted_model(	plot_dir		= fit_dir,
									case_tag		= sprintf('tree %d',n),
									model_name		= sprintf('HBDELCNeg - %s',fits$model_types[fits$best_model]),
									fsimulation		= fits$best_fit_sim,
									empirical_LTT	= LTT,
									root_age		= root_age,
									time_unit		= SIM_TIME_UNIT,
									only_class		= FALSE,
									verbose			= FALSE,
									verbose_prefix	= "      ")
				plot_fitted_vs_true_model(	plot_dir		= fit_dir,
											case_tag		= sprintf('tree %d',n),
											subtitle		= "Fitted BDELCNeg",
											true_model_name	= sprintf("sim %d",n),
											fit_model_name	= "BDELCNeg",
											sim_true		= sim_true,
											sim_fit			= fits$best_fit_sim,
											tree_LTT		= LTT,
											root_age		= root_age,
											time_unit		= SIM_TIME_UNIT,
											verbose			= FALSE,
											verbose_prefix	= "      ")
				# keep record of main EC fit results
				sim_results$BDELCNeg_type[n]			= fits$model_types[fits$best_model]
				sim_results$BDELCNeg_loglikelihood[n]	= fits$loglikelihoods[fits$best_model]
				sim_results$BDELCNeg_AIC[n]				= fits$AICs[fits$best_model]
				sim_results$BDELCNeg_lambda[n]			= fits$lambda0s[fits$best_model]
				sim_results$BDELCNeg_mu[n]				= fits$mu0s[fits$best_model]
				sim_results$BDELCNeg_minmu[n]			= fits$minmus[fits$best_model]
				sim_results$BDELCNeg_epsilon[n]			= fits$mu0s[fits$best_model]/fits$lambda0s[fits$best_model]
				sim_results$BDELCNeg_relative_error[n]	= (fits$mu0s[fits$best_model]-series_mu[1])/series_mu[1]
				sim_results$BDELCNeg_dLTT_R2[n]			= get_R2(xtrue=sim_true$ages, sim_true$LTT, xfit=fits$best_fit_sim$ages, yfit=fits$best_fit_sim$LTT, minx=0, maxx=root_age) # compare dLTT of fitted vs true model
			}


			# consider model congruent to the fitted BDELCNeg, but with the true mu
			cat2(sprintf("    Simulating and plotting model congruent to the fitted BDELCNeg model, with true mu..\n"))
			fsim_true_mu = simulate_deterministic_hbd(	LTT0			= Ntips,
														oldest_age		= root_age,
														rho0			= rho,
														lambda0			= fits$best_fit_sim$rholambda0/rho,
														age_grid		= fits$best_fit_sim$ages,
														mu				= approx(x=series_ages, y=series_mu, xout=fits$best_fit_sim$ages)$y,
														PDR				= fits$best_fit_sim$PDR,
														relative_dt		= INTEGRATION_RELATIVE_DT,
														allow_unreal	= FALSE)
			if(!fsim_true_mu$success){
				cat2(sprintf("      WARNING: Simulation failed: %s\n",fsim_true_mu$error))
			}else{
				congruent_dir = sprintf("%s/fitted_BDELCNeg_congruent_true_mu",this_sim_dir)
				plot_fitted_model(	plot_dir		= congruent_dir,
									case_tag		= sprintf('tree %d',n),
									model_name		= sprintf('HBDELCNeg - congruent, true mu'),
									fsimulation		= fsim_true_mu,
									empirical_LTT	= LTT,
									root_age		= root_age,
									time_unit		= SIM_TIME_UNIT,
									only_class		= FALSE,
									verbose			= FALSE,
									verbose_prefix	= "      ")
									
				# plot congruent lambda & mu vs true lambda & mu
				plot_curves(file_basepath 		= sprintf("%s/lambda_mu_vs_true",congruent_dir),
							xtype				= "age",
							ytype	 			= "speciation/extinction rates",
							case_tag 			= sprintf('tree %d',n),
							curves 				= list(	list(series_ages,series_lambda),
														list(series_ages,series_mu),
														list(fsim_true_mu$ages, fsim_true_mu$lambda),
														list(fsim_true_mu$ages, fsim_true_mu$mu)),
							curve_names 		= c("true lambda", "true mu", "fitted lambda", "fitted_mu"),
							curve_colors 		= c("black", "red", "black", "red"),
							curve_line_types	= c(1,1,2,2),
							curve_widths 		= c(1,2,1,2),
							reverse_x			= TRUE,
							plot_maxx 			= root_age,
							plot_miny			= 0,
							resolution			= PLOT_DOWNSAMPLING_RESOLUTION,
							xlabel 				= sprintf("age (%s)",SIM_TIME_UNIT),
							ylabel 				= sprintf("rate (1/%s)",SIM_TIME_UNIT),
							legend_pos 			= "outside",
							plot_title 			= sprintf("lambda & mu of fitted HBDELCNeg - congruent, true mu\ntree %d",n),
							data_file_comments 	= sprintf("# Speciation/extinction rates of fitted HBDELCNeg - congruent, true mu, tree %d",n),
							verbose 			= FALSE,
							verbose_prefix 		= "      ")
			}
		

			# fit ELC models, while fixing lambda to its true profile
			fit_dir = sprintf("%s/fitted_BDELC_true_lambda",this_sim_dir)
			cat2(sprintf("    Fitting ELC-TL models to simulated tree #%d, thus fixing lambda to its true profile..\n",n))
			tlfits = fit_ELC_models(tree			= tree, 
									rho				= rho,
									oldest_age		= fit_oldest_age,
									Ntrials			= FITTING_NTRIALS,
									negative_mu		= FALSE,
									fixed_lambda	= function(ages){ approx(x=series_ages, y=series_lambda, xout=ages)$y },
									verbose			= TRUE,
									verbose_prefix	= "      ")
			if(!tlfits$success){
				cat2(sprintf("      ERROR: Could not fit ELC-TL models: %s\n",tlfits$error))
			}else{
				dir.create(fit_dir, showWarnings = FALSE, recursive=TRUE);
				sink(file=sprintf("%s/fit_results.txt",fit_dir)); print(tlfits$model_fits[[tlfits$best_model]]); sink(); # save fit results to text file
				plot_fitted_vs_true_model(	plot_dir		= fit_dir,
											case_tag		= sprintf('tree %d',n),
											subtitle		= "Fitted BDELC-TL",
											true_model_name	= sprintf("sim %d",n),
											fit_model_name	= "BDELCTL",
											sim_true		= sim_true,
											sim_fit			= tlfits$best_fit_sim,
											tree_LTT		= LTT,
											root_age		= root_age,
											time_unit		= SIM_TIME_UNIT,
											verbose			= FALSE,
											verbose_prefix	= "      ")
				cat2(sprintf("    Assessing adequacy of fitted ELC-TL model..\n"))
				adequacy = assess_model_adequacy(	tree 				= tree, 
													model				= tlfits$best_fit_sim, 
													tree_name 			= tree_name, 
													model_name			= "BDELCTL_fit", 
													Nbootstraps			= MODEL_ADEQUACY_NBOOTSTRAPS, 
													report_file			= sprintf("%s/fit_model_adequacy.txt",fit_dir), 
													LTT_plot_basepath	= sprintf("%s/LTT_tree_vs_fitted_model",fit_dir),
													Nthreads			= NUMBER_OF_PARALLEL_THREADS)
				sim_results$BDELCTL_type[n]				= tlfits$model_types[tlfits$best_model]
				sim_results$BDELCTL_loglikelihood[n]	= tlfits$loglikelihoods[tlfits$best_model]
				sim_results$BDELCTL_AIC[n]				= tlfits$AICs[tlfits$best_model]
				sim_results$BDELCTL_mu[n]				= tlfits$mu0s[tlfits$best_model]
				sim_results$BDELCTL_PnodeKS[n]			= adequacy$PnodeKS
				sim_results$BDELCTL_relative_error[n]	= (tlfits$mu0s[tlfits$best_model]-series_mu[1])/series_mu[1]
			}
		}
		
		# free some RAM
		invisible(gc())
		n = n+1
	}

	cat2(sprintf("Saving results for all simulated trees in summary table\n"))
	output_table =sprintf("%s/results_summary.tsv",sim_dir)
	cat(sprintf("# Overview of simulated timetrees and BD fitting results\n#\n# %s\n",paste(colnames(sim_results),collapse="\t")), file=output_table, append=FALSE)
	write.table(x=sim_results, file=output_table, append=TRUE, sep="\t", row.names=FALSE, col.names=FALSE, quote=FALSE)

	pairwise_scatterplots(	output_basepath = sprintf("%s/BDELC_mu_vs_BDELCNeg_mu",sim_dir),
							case_tag		= sprintf("all simulations"),
							scattervalues	= cbind(BDELC	= sim_results$BDELC_mu,
													BDELCNeg	= sim_results$BDELCNeg_mu),
							data_names		= NULL,
							point_names		= sprintf("simulation %d",c(1:SIM_FITTING_NSIMULATIONS)),
							logarithmic		= FALSE,
							include_diagonal= FALSE,
							comment_line	= sprintf("Analysis of %d simulated timetrees, fitted present-day extinction rate (mu), BDELC vs BDELCNeg model",SIM_FITTING_NSIMULATIONS),
							verbose			= FALSE,
							verbose_prefix	= "  ")
							
	pairwise_scatterplots(	output_basepath = sprintf("%s/BDELC_minmu_vs_BDELCNeg_minmu",sim_dir),
							case_tag		= sprintf("all simulations"),
							scattervalues	= cbind(BDELC	= sim_results$BDELC_minmu,
													BDELCNeg	= sim_results$BDELCNeg_minmu),
							data_names		= NULL,
							point_names		= sprintf("simulation %d",c(1:SIM_FITTING_NSIMULATIONS)),
							logarithmic		= FALSE,
							include_diagonal= FALSE,
							comment_line	= sprintf("Analysis of %d simulated timetrees, fitted minimum extinction rate (mu), BDELC vs BDELCNeg model",SIM_FITTING_NSIMULATIONS),
							verbose			= FALSE,
							verbose_prefix	= "  ")

	pairwise_scatterplots(	output_basepath = sprintf("%s/BDELC_mu_vs_true_mu",sim_dir),
							case_tag		= sprintf("all simulations"),
							scattervalues	= cbind(fitted_mu	= sim_results$BDELC_mu,
													true_mu		= sim_results$mu0),
							data_names		= NULL,
							point_names		= sprintf("simulation %d",c(1:SIM_FITTING_NSIMULATIONS)),
							logarithmic		= FALSE,
							include_diagonal= FALSE,
							comment_line	= sprintf("Analysis of %d simulated timetrees, fitted BDELC present-day extinction rate (mu) vs. true present-day mu",SIM_FITTING_NSIMULATIONS),
							verbose			= FALSE,
							verbose_prefix	= "  ")
							
	pairwise_scatterplots(	output_basepath = sprintf("%s/BDELC_PnodeKS_vs_BDELC_mu",sim_dir),
							case_tag		= sprintf("all simulations"),
							scattervalues	= cbind(PnodeKS	= sim_results$BDELC_PnodeKS,
													mu		= sim_results$BDELC_mu),
							data_names		= NULL,
							point_names		= sprintf("simulation %d",c(1:SIM_FITTING_NSIMULATIONS)),
							logarithmic		= FALSE,
							include_diagonal= FALSE,
							comment_line	= sprintf("Pvalue of Kolomogorov-Smirnov test (node ages) and fitted BDELC_mu, for %d simulated trees",SIM_FITTING_NSIMULATIONS),
							verbose			= FALSE,
							verbose_prefix	= "  ")
}




if(INCLUDE_MODEL_FITTING_TO_REAL_TREES){
	real_trees_output_dir = sprintf("%s/real_trees",output_dir)
	cat2(sprintf("Loading tree specs table..\n"))
	tree_specs_table = read.table(	file 				= TREE_SPECS_FILE_PATH, 
									header 				= TRUE, 
									sep 				= "\t", 
									quote 				= "\"",
									dec					= ".",
									strip.white 		= TRUE, 
									blank.lines.skip 	= TRUE,
									na.strings 			= "",
									check.names			= FALSE,
									comment.char 		= "#",
									colClasses 			= NA,
									stringsAsFactors 	= FALSE)
	cat2(sprintf("  Note: Found %d trees in table\n",nrow(tree_specs_table)))

	# filter out some trees if required
	keep = filter_items_by_name(tree_specs_table[,'tree_name'], only_names=ONLY_TREES, omit_names=OMIT_TREES)
	if(length(keep)<nrow(tree_specs_table)){
		cat2(sprintf("  Note: Only considering %d out of %d trees, as requested\n", length(keep), nrow(tree_specs_table)))
		tree_specs_table = tree_specs_table[keep,]
	}

	# extract some tree specs and store as list of lists
	Ntrees = nrow(tree_specs_table)
	tree_specs = lapply(1:Ntrees, FUN=function(r) list(	name		= tree_specs_table[r,'tree_name'], 
														path 		= sprintf("%s/trees/%s.tre",dirname(TREE_SPECS_FILE_PATH),tree_specs_table[r,'tree_name']), 
														Nextant 	= tree_specs_table[r,'Nextant'],
														time_factor	= tree_specs_table[r,'time_factor'],
														time_unit	= tree_specs_table[r,'time_unit'],
														taxon		= tree_specs_table[r,'taxon']));


	# analyze trees
	real_results = data.frame(	Ntips_per_tree				= rep(NA, times=Ntrees),
								root_age					= rep(NA, times=Ntrees),
								BDELC_type		 			= character(Ntrees),
								BDELC_loglikelihood 		= rep(NA, times=Ntrees),
								BDELC_AIC					= rep(NA, times=Ntrees),
								BDELC_lambda				= rep(NA, times=Ntrees),
								BDELC_mu					= rep(NA, times=Ntrees),
								BDELC_minmu					= rep(NA, times=Ntrees),
								BDELC_epsilon				= rep(NA, times=Ntrees),
								BDELC_PColless				= rep(NA, times=Ntrees),
								BDELC_PSackin				= rep(NA, times=Ntrees),
								BDELC_PedgeKS				= rep(NA, times=Ntrees),
								BDELC_PnodeKS				= rep(NA, times=Ntrees),
								BDELC_fraction_LTT_in_CI95	= rep(NA, times=Ntrees),
								BDELC_LTT_R2				= rep(NA, times=Ntrees),
								BDELCNeg_type		 		= character(Ntrees),
								BDELCNeg_loglikelihood 		= rep(NA, times=Ntrees),
								BDELCNeg_AIC				= rep(NA, times=Ntrees),
								BDELCNeg_lambda				= rep(NA, times=Ntrees),
								BDELCNeg_mu					= rep(NA, times=Ntrees),
								BDELCNeg_minmu				= rep(NA, times=Ntrees),
								BDELCNeg_epsilon			= rep(NA, times=Ntrees),
								BDELCNeg_LTT_R2				= rep(NA, times=Ntrees))
	for(m in 1:Ntrees){
		external = tree_specs[[m]]
		cat2(sprintf("Examining empirical tree #%d (%s, %s)..\n",m,external$name,external$taxon));
	
		# load tree from file
		cat2(sprintf("  Loading tree from file..\n"))
		if(!file.exists(external$path)){
			cat2(sprintf("    WARNING: Missing input tree '%s'. Skipping tree\n",external$name))
			next
		}
		tree 			 = castor::read_tree(file=external$path);
		tree			 = castor::extend_tree_to_height(tree)$tree # enforce ultrametricity in case of rounding errors
		tree$edge.length = tree$edge.length * external$time_factor # rescale time units if needed
		Ntips 			 = length(tree$tip.label)
		root_age 		 = castor::get_tree_span(tree)$max_distance
		real_results$root_age[m] = root_age
		if(is.null(external$Nextant)) external$Nextant = Ntips;
		rarefaction		 = min(1,Ntips/external$Nextant);
		cat2(sprintf("    --> Loaded tree has %d tips and %d nodes, root_age=%g %s, rarefaction=%g\n",Ntips,tree$Nnode,root_age,external$time_unit,rarefaction));
		real_results$Ntips_per_tree[m] = Ntips
		if(Ntips<REAL_FITTIN_MIN_NTIPS){
			cat2(sprintf("  Note: Tree size (%d tips) is smaller than threshold (%d). Skipping tree\n",Ntips,REAL_FITTIN_MIN_NTIPS))
			next;
		}else if(rarefaction<REAL_FITTING_MIN_RAREFACTION){
			cat2(sprintf("  Note: Tree sampling fraction (%g) is smaller than threshold (%g). Skipping tree\n",rarefaction,REAL_FITTING_MIN_RAREFACTION))
			next;
		}

		# prepare tree-specific output dir
		tree_output_dir=sprintf("%s/results_per_tree/%s_%s",real_trees_output_dir,external$name,external$taxon);
		dir.create(tree_output_dir, showWarnings = FALSE, recursive=TRUE);

		# calculate empirical LTT	
		cat2(sprintf("  Calculating and plotting LTT..\n"))
		empirical_LTT = castor::count_lineages_through_time(tree,Ntimes=10*log2(Ntips));
		empirical_LTT$ages = root_age-empirical_LTT$times
		plot_curves(file_basepath 		= sprintf("%s/LTT",tree_output_dir),
					xtype				= "age",
					ytype	 			= "LTTs",
					case_tag 			= sprintf("%s",external$name),
					curves 				= list(	list(empirical_LTT$ages, empirical_LTT$lineages)),
					curve_names 		= c("LTT"),
					curve_colors 		= c(PLOT_COLOR_EMPIRICAL),
					curve_line_types	= PLOT_LINE_TYPE_EMPIRICAL,
					curve_widths 		= 2,
					reverse_x			= TRUE,
					plot_maxx 			= root_age,
					plot_miny			= 1,
					resolution			= PLOT_DOWNSAMPLING_RESOLUTION,
					xlabel 				= sprintf("age (%s)",external$time_unit),
					ylabel 				= "lineages",
					legend_pos 			= "outside",
					plot_title 			= sprintf("LTT of '%s' (%s)",external$name,external$taxon),
					data_file_comments 	= sprintf("LTT curve of external tree '%s', loaded from: %s",external$name,external$path),
					verbose 			= FALSE,
					verbose_prefix 		= "    ")

		# determine age span to consider
		oldest_age = (if(is.null(external$oldest_age)) root_age else external$oldest_age)
		oldest_age = min(oldest_age,root_age - empirical_LTT$times[which(empirical_LTT$lineages>=REAL_FITTING_MIN_LTT)[1]])
		if(oldest_age<root_age) cat2(sprintf("  Note: Considering only ages < %g %s for fitting\n", oldest_age,external$time_unit));
	
		best_BDELC_model_type = NA
		if(REAL_TREES_INCLUDE_HBD_ELC_FITTING){
			cat2(sprintf("  Fitting ELC models to tree '%s'..\n",external$name))
			fit_dir = sprintf("%s/fitted_BDELC",tree_output_dir)
			fits = fit_ELC_models(	tree			= tree, 
									rho				= rarefaction,
									oldest_age 		= oldest_age,
									Ntrials			= FITTING_NTRIALS,
									negative_mu		= FALSE,
									only_model_types= NULL,
									verbose			= TRUE,
									verbose_prefix	= "    ")
			if(!fits$success){
				cat2(sprintf("    ERROR: Could not fit ELC models: %s\n",fits$error))
			}else{
				cat2(sprintf("  Plotting fitted ELC model for tree '%s'..\n",external$name));
				dir.create(fit_dir, showWarnings = FALSE, recursive=TRUE);
				sink(file=sprintf("%s/fit_results.txt",fit_dir)); print(fits$model_fits[[fits$best_model]]); sink(); # save fit results to text file
				plot_fitted_model(	plot_dir		= fit_dir,
									case_tag		= sprintf('%s - %s',external$name,external$taxon),
									model_name		= sprintf('HBDELC - %s',fits$model_types[fits$best_model]),
									fsimulation		= fits$best_fit_sim,
									empirical_LTT	= empirical_LTT,
									root_age		= root_age,
									time_unit		= external$time_unit,
									only_class		= FALSE,
									verbose			= FALSE,
									verbose_prefix	= "      ")
				# assess adequacy of fitted model
				cat2(sprintf("  Assessing adequacy of ELC model fitted to tree '%s'..\n",external$name))
				adequacy = assess_model_adequacy(	tree 				= tree, 
													model				= fits$best_fit_sim, 
													tree_name 			= external$name, 
													model_name			= "BDELC_fit", 
													Nbootstraps			= MODEL_ADEQUACY_NBOOTSTRAPS, 
													report_file			= sprintf("%s/fit_model_adequacy.txt",fit_dir), 
													LTT_plot_basepath	= sprintf("%s/LTT_tree_vs_fitted_model",fit_dir),
													Nthreads			= NUMBER_OF_PARALLEL_THREADS)
				cat2(sprintf("   --> PedgeKS=%g, PnodeKS=%g, fraction LTT in CI95=%g\n",adequacy$PedgeKS,adequacy$PnodeKS,adequacy$fraction_LTT_in_CI95))
				# keep record of main EC fit results
				best_BDELC_model_type						= fits$model_types[fits$best_model]
				real_results$BDELC_type[m]					= fits$model_types[fits$best_model]
				real_results$BDELC_loglikelihood[m]			= fits$loglikelihoods[fits$best_model]
				real_results$BDELC_AIC[m]					= fits$AICs[fits$best_model]
				real_results$BDELC_lambda[m]				= fits$lambda0s[fits$best_model]
				real_results$BDELC_mu[m]					= fits$mu0s[fits$best_model]
				real_results$BDELC_minmu[m]					= fits$minmus[fits$best_model]
				real_results$BDELC_epsilon[m]				= fits$mu0s[fits$best_model]/fits$lambda0s[fits$best_model]
				real_results$BDELC_PColless[m]				= adequacy$PColless
				real_results$BDELC_PSackin[m]				= adequacy$PSackin
				real_results$BDELC_PedgeKS[m]				= adequacy$PedgeKS
				real_results$BDELC_PnodeKS[m]				= adequacy$PnodeKS
				real_results$BDELC_fraction_LTT_in_CI95[m]	= adequacy$fraction_LTT_in_CI95
				real_results$BDELC_LTT_R2[m]				= get_R2(xtrue=empirical_LTT$ages, ytrue=empirical_LTT$lineages, xfit=fits$best_fit_sim$ages, yfit=fits$best_fit_sim$LTT, minx=0, maxx=root_age) # compare dLTT of fitted vs empirical LTT

			}
		}
	
		if(REAL_TREES_INCLUDE_HBD_ELC_NEG_FITTING){
			cat2(sprintf("  Fitting ELC models (allowing negative mu) to tree '%s'..\n",external$name))
			fit_dir = sprintf("%s/fitted_BDELCNeg",tree_output_dir)
			fits = fit_ELC_models(	tree			= tree, 
									rho				= rarefaction,
									oldest_age 		= oldest_age,
									Ntrials			= FITTING_NTRIALS,
									negative_mu		= TRUE,
									only_model_types= (if(REAL_FITTING_BDELCNEG_USE_BDELC_MODEL && (!is.na(best_BDELC_model_type))) stringr::str_replace_all(best_BDELC_model_type,"mu_zero","mu_const") else NULL),
									param_guesses	= (if(REAL_FITTING_BDELCNEG_USE_BDELC_AS_GUESS && fits$success) fits$fitted_params else NULL),
									verbose			= TRUE,
									verbose_prefix	= "    ")
			if(!fits$success){
				cat2(sprintf("    ERROR: Could not fit ELC models: %s\n",fits$error))
			}else{
				cat2(sprintf("  Plotting fitted ELC model for tree '%s'..\n",external$name));
				dir.create(fit_dir, showWarnings = FALSE, recursive=TRUE);
				sink(file=sprintf("%s/fit_results.txt",fit_dir)); print(fits$model_fits[[fits$best_model]]); sink(); # save fit results to text file
				plot_fitted_model(	plot_dir		= fit_dir,
									case_tag		= sprintf('%s - %s',external$name,external$taxon),
									model_name		= sprintf('HBDELCNeg - %s',fits$model_types[fits$best_model]),
									fsimulation		= fits$best_fit_sim,
									empirical_LTT	= empirical_LTT,
									root_age		= root_age,
									time_unit		= external$time_unit,
									only_class		= FALSE,
									verbose			= FALSE,
									verbose_prefix	= "      ")
				# keep record of main EC fit results
				real_results$BDELCNeg_type[m]			= fits$model_types[fits$best_model]
				real_results$BDELCNeg_loglikelihood[m]	= fits$loglikelihoods[fits$best_model]
				real_results$BDELCNeg_AIC[m]			= fits$AICs[fits$best_model]
				real_results$BDELCNeg_lambda[m]			= fits$lambda0s[fits$best_model]
				real_results$BDELCNeg_mu[m]				= fits$mu0s[fits$best_model]
				real_results$BDELCNeg_minmu[m]			= fits$minmus[fits$best_model]
				real_results$BDELCNeg_epsilon[m]		= fits$mu0s[fits$best_model]/fits$lambda0s[fits$best_model]
				real_results$BDELCNeg_LTT_R2[m]			= get_R2(xtrue=empirical_LTT$ages, ytrue=empirical_LTT$lineages, xfit=fits$best_fit_sim$ages, yfit=fits$best_fit_sim$LTT, minx=0, maxx=root_age) # compare dLTT of fitted vs empirical LTT
			}
		}

		# free some RAM
		invisible(gc())
	}


	# plot results across all trees
	cat2(sprintf("Plotting results across all trees..\n"))
	pairwise_scatterplots(	output_basepath = sprintf("%s/all_trees/BDELC_mu_vs_BDELCNeg_mu",real_trees_output_dir),
							case_tag		= sprintf("fitted mu0, %d trees",Ntrees),
							scattervalues	= cbind(BDELC		= real_results$BDELC_mu,
													BDELCNeg	= real_results$BDELCNeg_mu),
							data_names		= NULL,
							point_names		= tree_specs_table[,'tree_name'],
							logarithmic		= FALSE,
							include_diagonal= FALSE,
							comment_line	= sprintf("Analysis of %d timetrees, fitted present-day extinction rate (mu), BDELC vs BDELCNeg model",Ntrees),
							verbose			= FALSE,
							verbose_prefix	= "  ")

	pairwise_scatterplots(	output_basepath = sprintf("%s/all_trees/BDELC_minmu_vs_BDELCNeg_minmu",real_trees_output_dir),
							case_tag		= sprintf("fitted minmu, %d trees",Ntrees),
							scattervalues	= cbind(BDELC		= real_results$BDELC_minmu,
													BDELCNeg	= real_results$BDELCNeg_minmu),
							data_names		= NULL,
							point_names		= tree_specs_table[,'tree_name'],
							logarithmic		= FALSE,
							include_diagonal= FALSE,
							comment_line	= sprintf("Analysis of %d timetrees, fitted minimum extinction rate (mu), BDELC vs BDELCNeg model",Ntrees),
							verbose			= FALSE,
							verbose_prefix	= "  ")

	pairwise_scatterplots(	output_basepath = sprintf("%s/all_trees/BDELC_statistical_test_Pvalues",real_trees_output_dir),
							case_tag		= sprintf("%d trees",Ntrees),
							scattervalues	= cbind(Colless	= real_results$BDELC_PColless,
													Sackin	= real_results$BDELC_PSackin,
													edgeKS	= real_results$BDELC_PedgeKS,
													nodeKS	= real_results$BDELC_PnodeKS),
							data_names		= NULL,
							point_names		= tree_specs_table[,'tree_name'],
							logarithmic		= FALSE,
							include_diagonal= FALSE,
							comment_line	= sprintf("Statistical significances of various tests for %d timetrees under the BDELC model",Ntrees),
							verbose			= FALSE,
							verbose_prefix	= "  ")
					
	pairwise_scatterplots(	output_basepath = sprintf("%s/all_trees/BDELC_PedgeKS_vs_mu",real_trees_output_dir),
							case_tag		= sprintf("%d trees",Ntrees),
							scattervalues	= cbind(edgeKS	= real_results$BDELC_PedgeKS,
													mu		= real_results$BDELC_mu),
							data_names		= NULL,
							point_names		= tree_specs_table[,'tree_name'],
							logarithmic		= FALSE,
							include_diagonal= FALSE,
							comment_line	= sprintf("Statistical significance of Kolmogorov-Smirnov test (edge lengths) vs fitted present-day mu, for %d timetrees under the BDELC model",Ntrees),
							verbose			= FALSE,
							verbose_prefix	= "  ")
					
	pairwise_scatterplots(	output_basepath = sprintf("%s/all_trees/BDELC_PnodeKS_vs_mu",real_trees_output_dir),
							case_tag		= sprintf("%d trees",Ntrees),
							scattervalues	= cbind(nodeKS	= real_results$BDELC_PnodeKS,
													mu		= real_results$BDELC_mu),
							data_names		= NULL,
							point_names		= tree_specs_table[,'tree_name'],
							logarithmic		= FALSE,
							include_diagonal= FALSE,
							comment_line	= sprintf("Statistical significance of Kolmogorov-Smirnov test (node lengths) vs fitted present-day mu, for %d timetrees under the BDELC model",Ntrees),
							verbose			= FALSE,
							verbose_prefix	= "  ")

	# add summary results to tree_specs table and save table to file
	tree_specs_table = cbind(tree_specs_table, real_results)
	cat2(sprintf("Saving results for all analyzed trees in summary table\n"))
	output_table =sprintf("%s/results_summary.tsv",real_trees_output_dir)
	check_output_file(output_table,TRUE,TRUE,"  ")
	cat(sprintf("# Overview of timetrees and BD fitting results\n#\n# %s\n",paste(colnames(tree_specs_table),collapse="\t")), file=output_table, append=FALSE)
	write.table(x=tree_specs_table, file=output_table, append=TRUE, sep="\t", row.names=FALSE, col.names=FALSE, quote=FALSE);
}


cat2(sprintf("Done. All outputs were written to '%s'\n",output_dir));
